from typing import Union, Any
from mutf8 import encode_modified_utf8
from ._dtype import AnyNBT, EncoderType

class AbstractBase:
    def to_snbt(self, indent: Union[int, str] = None) -> str: ...
    def to_nbt(
        self,
        *,
        compressed=True,
        little_endian=False,
        string_encoder: EncoderType = encode_modified_utf8,
    ) -> bytes: ...
    def save_to(
        self,
        filepath_or_buffer=None,
        *,
        compressed=True,
        little_endian=False,
        string_encoder: EncoderType = encode_modified_utf8,
    ) -> bytes: ...

class AbstractBaseTag(AbstractBase):
    tag_id: int
    @property
    def py_data(self) -> Any: ...
    def to_nbt(
        self,
        *,
        compressed=True,
        little_endian=False,
        string_encoder: EncoderType = encode_modified_utf8,
        name="",
    ) -> bytes: ...
    def save_to(
        self,
        filepath_or_buffer=None,
        *,
        compressed=True,
        little_endian=False,
        string_encoder: EncoderType = encode_modified_utf8,
        name="",
    ) -> bytes: ...
    def __repr__(self) -> str:
        raise NotImplementedError
    def __str__(self) -> str:
        raise NotImplementedError
    def __eq__(self, other) -> bool:
        raise NotImplementedError
    def __reduce__(self):
        raise NotImplementedError
    def copy(self) -> AnyNBT: ...
    def __deepcopy__(self, memo=None):
        raise NotImplementedError
    def __copy__(self):
        raise NotImplementedError

class AbstractBaseImmutableTag(AbstractBaseTag):
    def __hash__(self) -> int:
        raise NotImplementedError

class AbstractBaseMutableTag(AbstractBaseTag):
    __hash__: None



from typing import Optional

class NBTError(Exception):
    """Some error in the NBT library."""


class NBTLoadError(NBTError):
    """The NBT data failed to load for some reason."""
    pass


class NBTFormatError(NBTLoadError):
    """Indicates the NBT format is invalid."""
    pass


class SNBTParseError(NBTError):
    """Indicates the SNBT format is invalid."""

    index: Optional[int]  # The index at which the error occurred

    def __init__(self, *args, index=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.index = index



import re

from ._errors import SNBTParseError
from ._value cimport AbstractBaseTag
from ._int cimport ByteTag, ShortTag, IntTag, LongTag
from ._float cimport FloatTag, DoubleTag
from ._array cimport ByteArrayTag, IntArrayTag, LongArrayTag
from ._string cimport StringTag
from ._list cimport CyListTag
from ._compound cimport CyCompoundTag

whitespace = re.compile(r"[ \t\r\n]*")
int_numeric = re.compile(r"[+-]?\d+[bBsSlL]?")
# If the letter code and decimal do not exist it is parsed as a string tag.
float_numeric = re.compile(r"[+-]?((((\d+\.\d*)|(\d*\.\d+))([eE][+-]?\d+)?)|((\d+|(\d+\.\d*)|(\d*\.\d+))([eE][+-]?\d+)?[fFdD]))")
alnumplus = re.compile(r"[A-Za-z0-9._+-]+")
comma = re.compile(r"[ \t\r\n]*,[ \t\r\n]*")
colon = re.compile(r"[ \t\r\n]*:[ \t\r\n]*")
array_lookup = {'B': ByteArrayTag, 'I': IntArrayTag, 'L': LongArrayTag}

cdef int _strip_whitespace(unicode snbt, int index):
    cdef object match

    match = whitespace.match(snbt, index)
    if match is None:
        return index
    else:
        return match.end()

cdef int _strip_comma(unicode snbt, int index, unicode end_chr) except -1:
    cdef object match

    match = comma.match(snbt, index)
    if match is None:
        index = _strip_whitespace(snbt, index)
        if snbt[index] != end_chr:
            raise SNBTParseError(
                f'Expected a comma or {end_chr} at {index} but got ->{snbt[index:index + 10]} instead',
                index=index
            )
    else:
        index = match.end()
    return index

cdef int _strip_colon(unicode snbt, int index) except -1:
    cdef object match

    match = colon.match(snbt, index)
    if match is None:
        raise SNBTParseError(
            f'Expected : at {index} but got ->{snbt[index:index + 10]} instead',
            index=index
        )
    else:
        return match.end()

cdef inline unescape(unicode string):
    return string.replace('\\"', '"').replace("\\\\", "\\")

cdef tuple _capture_string(unicode snbt, int index):
    cdef unicode val, quote
    cdef bint strict_str
    cdef int end_index
    cdef object match

    if snbt[index] in ('"', "'"):
        quote = snbt[index]
        strict_str = True
        index += 1
        end_index = index
        while not (snbt[end_index] == quote and not (len(snbt[:end_index]) - len(snbt[:end_index].rstrip('\\')))):
            end_index += 1
        val = unescape(snbt[index:end_index])
        index = end_index + 1
    else:
        strict_str = False
        match = alnumplus.match(snbt, index)
        if match is None:
            raise SNBTParseError(
                f'Expected a string at {index} but got ->{snbt[index:index + 10]} instead',
                index=index
            )
        val = match.group()
        index = match.end()

    return val, strict_str, index

cdef tuple _parse_snbt_recursive(unicode snbt, int index=0):
    cdef dict data_
    cdef list array
    cdef unicode array_type_chr
    cdef object array_type, first_data_type
    cdef AbstractBaseTag nested_data, data
    cdef unicode val
    cdef bint strict_str

    index = _strip_whitespace(snbt, index)
    if snbt[index] == '{':
        data_ = {}
        index += 1
        index = _strip_whitespace(snbt, index)
        while snbt[index] != '}':
            # read the key
            key, _, index = _capture_string(snbt, index)

            # get around the colon
            index = _strip_colon(snbt, index)

            # load the data and save it to the dictionary
            nested_data, index = _parse_snbt_recursive(snbt, index)
            data_[key] = nested_data

            index = _strip_comma(snbt, index, '}')
        data = CyCompoundTag.create(data_)
        # skip the }
        index += 1

    elif snbt[index] == '[':
        index += 1
        index = _strip_whitespace(snbt, index)
        if snbt[index:index + 2] in {'B;', 'I;', 'L;'}:
            # array
            array = []
            array_type_chr = snbt[index]
            array_type = array_lookup[array_type_chr]
            index += 2
            index = _strip_whitespace(snbt, index)

            while snbt[index] != ']':
                match = int_numeric.match(snbt, index)
                if match is None:
                    raise SNBTParseError(
                        f'Expected an integer value or ] at {index} but got ->{snbt[index:index + 10]} instead',
                        index=index
                    )
                else:
                    val = match.group()
                    if array_type_chr in {"B", "L"}:
                        if val[-1].upper() == array_type_chr:
                            val = val[:-1]
                        else:
                            raise SNBTParseError(
                                f'Expected the datatype marker "{array_type_chr}" at {index} but got ->{snbt[index:index + 10]} instead',
                                index=index
                            )
                    array.append(int(val))
                    index = match.end()

                index = _strip_comma(snbt, index, ']')
            data = array_type(array)
        else:
            # list
            array = []
            first_data_type = None
            while snbt[index] != ']':
                nested_data, index_ = _parse_snbt_recursive(snbt, index)
                if first_data_type is None:
                    first_data_type = nested_data.__class__
                if not isinstance(nested_data, first_data_type):
                    raise SNBTParseError(
                        f'Expected type {first_data_type.__name__} but got {nested_data.__class__.__name__} at {index}',
                        index=index
                    )
                else:
                    index = index_
                array.append(nested_data)
                index = _strip_comma(snbt, index, ']')

            if first_data_type is None:
                data = CyListTag.create()
            else:
                data = CyListTag.create(array, first_data_type().tag_id)

        # skip the ]
        index += 1

    else:
        val, strict_str, index = _capture_string(snbt, index)
        if strict_str:
            data = StringTag(val)
        else:
            if int_numeric.fullmatch(val) is not None:
                # we have an int type
                if val[-1] in {'b', 'B'}:
                    data = ByteTag(int(val[:-1]))
                elif val[-1] in {'s', 'S'}:
                    data = ShortTag(int(val[:-1]))
                elif val[-1] in {'l', 'L'}:
                    data = LongTag(int(val[:-1]))
                else:
                    data = IntTag(int(val))
            elif float_numeric.fullmatch(val) is not None:
                # we have a float type
                if val[-1] in {'f', 'F'}:
                    data = FloatTag(val[:-1])
                elif val[-1] in {'d', 'D'}:
                    data = DoubleTag(val[:-1])
                else:
                    data = DoubleTag(val)
            elif val.lower() == "false":
                data = ByteTag(0)
            elif val.lower() == "true":
                data = ByteTag(1)
            else:
                # we just have a string type
                data = StringTag(val)

    return data, index

cpdef AbstractBaseTag from_snbt(unicode snbt):
    try:
        return _parse_snbt_recursive(snbt)[0]
    except SNBTParseError as e:
        raise e
    except IndexError:
        raise SNBTParseError('SNBT string is incomplete. Reached the end of the string.')
    except Exception as e:
        raise SNBTParseError(e)



from typing import SupportsFloat
from ._numeric import AbstractBaseNumericTag

class AbstractBaseFloatTag(AbstractBaseNumericTag):
    def __init__(self, value: SupportsFloat = 0): ...
    @property
    def py_float(self) -> float: ...

class FloatTag(AbstractBaseFloatTag): ...
class DoubleTag(AbstractBaseFloatTag): ...



from cpython.bytes cimport PyBytes_FromStringAndSize, PyBytes_AsString
from libc.stdlib cimport malloc, free
from libc.string cimport memcpy
import codecs
import re

from . import __major__
from ._value cimport AbstractBaseTag
from ._errors import NBTFormatError
from ._dtype import DecoderType, EncoderType


cdef class BufferContext:
    def __cinit__(self):
        self.buffer = NULL
        self.size = 0
        self.offset = 0

    def __init__(
        self,
        bytes buffer = b"",
        size_t offset = 0
    ):
        self.size = len(buffer)
        self.buffer = <char*> malloc(self.size * sizeof(char))
        memcpy(self.buffer, PyBytes_AsString(buffer), self.size)
        if 0 <= offset <= self.size:
            self.offset = offset
        else:
            raise ValueError("offset must be between 0 and the length of the buffer.")

    cpdef bytes get_buffer(self):
        return PyBytes_FromStringAndSize(self.buffer, self.size)

    def __dealloc__(self):
        if self.buffer is not NULL:
            free(self.buffer)


cdef char *read_data(BufferContext buffer, size_t tag_size) except NULL:
    cdef char *value = buffer.buffer + buffer.offset
    buffer.offset += tag_size
    if buffer.offset > buffer.size:
        raise NBTFormatError(
            f"NBT Stream too short. Asked for {tag_size:d} bytes, {buffer.offset - buffer.size:d} more than the buffer."
        )
    return value

cdef void to_little_endian(void *data_buffer, int num_bytes, bint little_endian = False):
    if little_endian:
        return

    cdef unsigned char *buf = <unsigned char *> data_buffer
    cdef int i

    for i in range((num_bytes + 1) // 2):
        buf[i], buf[num_bytes - i - 1] = buf[num_bytes - i - 1], buf[i]


cdef inline char read_byte(BufferContext buffer) except? -1:
    return read_data(buffer, 1)[0]


cdef inline int read_int(BufferContext buffer, bint little_endian) except? -1:
    cdef int*pointer = <int*> read_data(buffer, 4)
    cdef int value = pointer[0]
    to_little_endian(&value, 4, little_endian)
    return value


cdef inline str read_string(BufferContext buffer, bint little_endian, string_decoder: DecoderType):
    return string_decoder(read_bytes(buffer, little_endian))

cdef inline bytes read_bytes(BufferContext buffer, bint little_endian):
    cdef unsigned short *pointer = <unsigned short *> read_data(buffer, 2)
    cdef unsigned short length = pointer[0]
    to_little_endian(&length, 2, little_endian)
    cdef char *c = read_data(buffer, length)
    return PyBytes_FromStringAndSize(c, length)

cdef inline void cwrite(object obj, char*buf, size_t length):
    obj.write(buf[:length])

cdef inline void write_string(str s, object buffer, bint little_endian, string_encoder: EncoderType):
    cdef bytes b = string_encoder(s)
    write_bytes(b, buffer, little_endian)

cdef inline void write_bytes(bytes b, object buffer, bint little_endian):
    cdef char *c = b
    cdef short length = <short> len(b)
    write_short(length, buffer, little_endian)
    cwrite(buffer, c, length)

cdef inline void write_array(object value, object buffer, char size, bint little_endian):
    value = value.tobytes()
    cdef char*s = value
    cdef int length = <int> len(value) // size
    to_little_endian(&length, 4, little_endian)
    cwrite(buffer, <char*> &length, 4)
    cwrite(buffer, s, len(value))

cdef inline void write_byte(char value, object buffer):
    cwrite(buffer, &value, 1)

cdef inline void write_short(short value, object buffer, bint little_endian):
    to_little_endian(&value, 2, little_endian)
    cwrite(buffer, <char*> &value, 2)

cdef inline void write_int(int value, object buffer, bint little_endian):
    to_little_endian(&value, 4, little_endian)
    cwrite(buffer, <char*> &value, 4)

cdef inline void write_long(long long value, object buffer, bint little_endian):
    to_little_endian(&value, 8, little_endian)
    cwrite(buffer, <char*> &value, 8)

cdef inline void write_float(float value, object buffer, bint little_endian):
    to_little_endian(&value, 4, little_endian)
    cwrite(buffer, <char*> &value, 4)

cdef inline void write_double(double value, object buffer, bint little_endian):
    to_little_endian(&value, 8, little_endian)
    cwrite(buffer, <char*> &value, 8)


# Functions for string encoding

cpdef str utf8_decoder(bytes b):
    """Standard UTF-8 decoder"""
    return b.decode()


cpdef bytes utf8_encoder(str s):
    """Standard UTF-8 encoder"""
    return s.encode()


def _escape_replace(err):
    if isinstance(err, UnicodeDecodeError):
        return f"␛x{err.object[err.start]:02x}", err.start+1
    raise err


codecs.register_error("escapereplace", _escape_replace)


cpdef str utf8_escape_decoder(bytes b):
    """UTF-8 decoder that escapes error bytes to the form ␛xFF"""
    return b.decode(errors="escapereplace")

EscapePattern = re.compile(b"\xe2\x90\x9bx([0-9a-zA-Z]{2})")  # ␛xFF

cpdef bytes _utf8_unescape(object m):
    return bytes([int(m.groups()[0], 16)])

cpdef bytes utf8_escape_encoder(str s):
    """UTF-8 encoder that converts ␛x[0-9a-fA-F]{2} back to individual bytes"""
    return EscapePattern.sub(_utf8_unescape, s.encode())

if __major__ <= 2:
    def primitive_conversion(obj):
        return obj.py_data if isinstance(obj, AbstractBaseTag) else obj



from ._util cimport BufferContext
from ._value cimport AbstractBaseTag
from ._dtype import DecoderType

cdef AbstractBaseTag load_payload(BufferContext buffer, char tag_type, bint little_endian, string_decoder: DecoderType)
cdef tuple load_tag(BufferContext buffer, bint little_endian, string_decoder: DecoderType)

cdef class ReadContext:
    cdef readonly int offset



from ._value cimport AbstractBaseImmutableTag
from ._util cimport BufferContext
from ._dtype import DecoderType


cdef StringTag read_string_tag(BufferContext buffer, bint little_endian, string_decoder: DecoderType)

cdef class StringTag(AbstractBaseImmutableTag):
    cdef str value_



from typing import Union, BinaryIO, List
from ._util import DecoderType
from mutf8 import decode_modified_utf8
from ._named_tag import NamedTag

class ReadContext:
    offset: int

def load(
    filepath_or_buffer: Union[str, bytes, BinaryIO, memoryview, None],
    *,
    compressed: bool = True,
    little_endian: bool = False,
    read_context: ReadContext = None,
    string_decoder: DecoderType = decode_modified_utf8,
) -> NamedTag: ...
def load_many(
    filepath_or_buffer: Union[str, bytes, BinaryIO, memoryview, None],
    *,
    count: int = 1,
    compressed: bool = True,
    little_endian: bool = False,
    read_context: ReadContext = None,
    string_decoder: DecoderType = decode_modified_utf8,
) -> List[NamedTag]: ...



from ._value import AbstractBaseMutableTag
import numpy

class AbstractBaseArrayTag(AbstractBaseMutableTag):
    def __init__(self, value=()): ...
    def __getitem__(self, item): ...
    def __setitem__(self, item): ...
    def __array__(self, dtype=None): ...
    def __len__(self): ...
    @property
    def np_array(self) -> numpy.ndarray: ...

class ByteArrayTag(AbstractBaseArrayTag): ...
class IntArrayTag(AbstractBaseArrayTag): ...
class LongArrayTag(AbstractBaseArrayTag): ...



cdef str CommaNewline
cdef str CommaSpace

cdef char ID_END
cdef char ID_BYTE
cdef char ID_SHORT
cdef char ID_INT
cdef char ID_LONG
cdef char ID_FLOAT
cdef char ID_DOUBLE
cdef char ID_BYTE_ARRAY
cdef char ID_STRING
cdef char ID_LIST
cdef char ID_COMPOUND
cdef char ID_INT_ARRAY
cdef char ID_LONG_ARRAY
cdef char ID_MAX



from typing import Union, Callable, TYPE_CHECKING

if TYPE_CHECKING:
    from ._int import (
        ByteTag,
        ShortTag,
        IntTag,
        LongTag,
    )
    from ._float import (
        FloatTag,
        DoubleTag,
    )
    from ._array import (
        ByteArrayTag,
        IntArrayTag,
        LongArrayTag,
    )
    from ._string import StringTag
    from ._list import ListTag
    from ._compound import CompoundTag

SNBTType = str

IntType = Union["ByteTag", "ShortTag", "IntTag", "LongTag"]

FloatType = Union["FloatTag", "DoubleTag"]

NumberType = Union["ByteTag", "ShortTag", "IntTag", "LongTag", "FloatTag", "DoubleTag"]

ArrayType = Union["ByteArrayTag", "IntArrayTag", "LongArrayTag"]

AnyNBT = Union[
    "ByteTag",
    "ShortTag",
    "IntTag",
    "LongTag",
    "FloatTag",
    "DoubleTag",
    "ByteArrayTag",
    "StringTag",
    "ListTag",
    "CompoundTag",
    "IntArrayTag",
    "LongArrayTag",
]

EncoderType = Callable[[str], bytes]
DecoderType = Callable[[bytes], str]



from typing import SupportsInt
from ._numeric import AbstractBaseNumericTag

class AbstractBaseIntTag(AbstractBaseNumericTag):
    def __init__(self, value: SupportsInt = 0): ...
    @property
    def py_int(self) -> int: ...

class ByteTag(AbstractBaseIntTag): ...
class ShortTag(AbstractBaseIntTag): ...
class IntTag(AbstractBaseIntTag): ...
class LongTag(AbstractBaseIntTag): ...



# Base Types
from ._value cimport AbstractBaseTag, AbstractBaseImmutableTag, AbstractBaseMutableTag
from ._numeric cimport AbstractBaseNumericTag

# Types
from ._int cimport (
    AbstractBaseIntTag,
    ByteTag,
    ShortTag,
    IntTag,
    LongTag,
)
from ._float cimport (
    AbstractBaseFloatTag,
    FloatTag,
    DoubleTag,
)
from ._array cimport (
    AbstractBaseArrayTag,
    ByteArrayTag,
    IntArrayTag,
    LongArrayTag,
)
from ._string cimport (
    StringTag,
)
from ._list cimport (
    CyListTag,
)
from ._compound cimport (
    CyCompoundTag,
)

from ._named_tag cimport NamedTag

# Load functions
from ._load_snbt cimport from_snbt



## This file is generated by tempita. Do not modify this file directly or your changes will get overwritten.
## To edit this file edit the template in template/src

from io import BytesIO
from copy import deepcopy
import warnings

from . import __major__
from ._numeric cimport AbstractBaseNumericTag
from ._const cimport ID_BYTE, ID_SHORT, ID_INT, ID_LONG
from ._util cimport (
    write_byte,
    write_short,
    write_int,
    write_long,
    BufferContext,
    read_data,
    to_little_endian,
    read_byte,
)
if __major__ <= 2:
    from ._util import primitive_conversion
from ._dtype import EncoderType


cdef class AbstractBaseIntTag(AbstractBaseNumericTag):
    """Abstract Base Class for all int Tag classes"""

    @property
    def py_int(AbstractBaseNumericTag self) -> int:
        """
        A python int representation of the class.
        The returned data is immutable so changes will not mirror the instance.
        """
        raise NotImplementedError

    @property
    def py_data(self):
        """
        A python representation of the class. Note that the return type is undefined and may change in the future.
        You would be better off using the py_{type} or np_array properties if you require a fixed type.
        This is here for convenience to get a python representation under the same property name.
        """
        return self.py_int


cdef inline ByteTag read_byte_tag(BufferContext buffer, bint little_endian):
    cdef ByteTag tag = ByteTag.__new__(ByteTag)
    tag.value_ = read_byte(buffer)
    return tag


cdef class ByteTag(AbstractBaseIntTag):
    """
    A 1 byte integer class.
    Can Store numbers between -(2^7) and (2^7 - 1)
    """
    tag_id = ID_BYTE

    def __init__(ByteTag self, value = 0):
        self.value_ = self._sanitise_value(int(value))

    def __str__(ByteTag self):
        return str(self.value_)

    def __eq__(ByteTag self, other):
        cdef ByteTag other_
        if isinstance(other, ByteTag):
            other_ = other
            return self.value_ == other_.value_
        elif __major__ <= 2:
            warnings.warn("NBT comparison operator (a == b) will only return True between classes of the same type.", FutureWarning)
            return self.value_ == primitive_conversion(other)
        return NotImplemented

    def __reduce__(ByteTag self):
        return self.__class__, (self.value_,)

    def __deepcopy__(ByteTag self, memo=None):
        return self.__class__(deepcopy(self.value_, memo=memo))

    def __copy__(ByteTag self):
        return self.__class__(self.value_)

    def __hash__(ByteTag self):
        return hash((self.tag_id, self.value_))

    def __ge__(ByteTag self, other):
        cdef ByteTag other_
        if isinstance(other, ByteTag):
            other_ = other
            return self.value_ >= other_.value_
        elif __major__ <= 2:
            warnings.warn("NBT comparison operator (a >= b) will only return True between classes of the same type.", FutureWarning)
            return self.value_ >= primitive_conversion(other)
        return NotImplemented

    def __gt__(ByteTag self, other):
        cdef ByteTag other_
        if isinstance(other, ByteTag):
            other_ = other
            return self.value_ > other_.value_
        elif __major__ <= 2:
            warnings.warn("NBT comparison operator (a > b) will only return True between classes of the same type.", FutureWarning)
            return self.value_ > primitive_conversion(other)
        return NotImplemented

    def __le__(ByteTag self, other):
        cdef ByteTag other_
        if isinstance(other, ByteTag):
            other_ = other
            return self.value_ <= other_.value_
        elif __major__ <= 2:
            warnings.warn("NBT comparison operator (a <= b) will only return True between classes of the same type.", FutureWarning)
            return self.value_ <= primitive_conversion(other)
        return NotImplemented

    def __lt__(ByteTag self, other):
        cdef ByteTag other_
        if isinstance(other, ByteTag):
            other_ = other
            return self.value_ < other_.value_
        elif __major__ <= 2:
            warnings.warn("NBT comparison operator (a == b) will only return True between classes of the same type.", FutureWarning)
            return self.value_ < primitive_conversion(other)
        return NotImplemented

    def __repr__(ByteTag self):
        return f"{self.__class__.__name__}({self.value_})"

    def __int__(ByteTag self):
        return self.value_.__int__()

    def __float__(ByteTag self):
        return self.value_.__float__()

    def __bool__(ByteTag self):
        return self.value_.__bool__()

    @property
    def py_int(ByteTag self) -> int:
        """
        A python int representation of the class.
        The returned data is immutable so changes will not mirror the instance.
        """
        return self.value_

    if __major__ <= 2:
        def __add__(self, other):
            warnings.warn(f"__add__ is depreciated on ByteTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) + primitive_conversion(other)

        def __sub__(self, other):
            warnings.warn(f"__sub__ is depreciated on ByteTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) - primitive_conversion(other)

        def __mul__(self, other):
            warnings.warn(f"__mul__ is depreciated on ByteTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) * primitive_conversion(other)

        def __truediv__(self, other):
            warnings.warn(f"__truediv__ is depreciated on ByteTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) / primitive_conversion(other)

        def __floordiv__(self, other):
            warnings.warn(f"__floordiv__ is depreciated on ByteTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) // primitive_conversion(other)

        def __mod__(self, other):
            warnings.warn(f"__mod__ is depreciated on ByteTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) % primitive_conversion(other)

        def __divmod__(self, other):
            warnings.warn(f"__divmod__ is depreciated on ByteTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return divmod(primitive_conversion(self), primitive_conversion(other))

        def __pow__(self, power, modulo):
            warnings.warn(f"__pow__ is depreciated on ByteTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return pow(primitive_conversion(self), power, modulo)

        def __lshift__(self, other):
            warnings.warn(f"__lshift__ is depreciated on ByteTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) << primitive_conversion(other)

        def __rshift__(self, other):
            warnings.warn(f"__rshift__ is depreciated on ByteTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) >> primitive_conversion(other)

        def __and__(self, other):
            warnings.warn(f"__and__ is depreciated on ByteTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) & primitive_conversion(other)

        def __xor__(self, other):
            warnings.warn(f"__xor__ is depreciated on ByteTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) ^ primitive_conversion(other)

        def __or__(self, other):
            warnings.warn(f"__or__ is depreciated on ByteTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) | primitive_conversion(other)

        def __radd__(self, other):
            warnings.warn(f"__radd__ is depreciated on ByteTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) + primitive_conversion(self)

        def __rsub__(self, other):
            warnings.warn(f"__rsub__ is depreciated on ByteTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) - primitive_conversion(self)

        def __rmul__(self, other):
            warnings.warn(f"__rmul__ is depreciated on ByteTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) * primitive_conversion(self)

        def __rtruediv__(self, other):
            warnings.warn(f"__rtruediv__ is depreciated on ByteTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) / primitive_conversion(self)

        def __rfloordiv__(self, other):
            warnings.warn(f"__rfloordiv__ is depreciated on ByteTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) // primitive_conversion(self)

        def __rmod__(self, other):
            warnings.warn(f"__rmod__ is depreciated on ByteTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) % primitive_conversion(self)

        def __rdivmod__(self, other):
            warnings.warn(f"__rdivmod__ is depreciated on ByteTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return divmod(primitive_conversion(other), primitive_conversion(self))

        def __rpow__(self, other, modulo):
            warnings.warn(f"__rpow__ is depreciated on ByteTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return pow(primitive_conversion(other), primitive_conversion(self), modulo)

        def __rlshift__(self, other):
            warnings.warn(f"__rlshift__ is depreciated on ByteTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) << primitive_conversion(self)

        def __rrshift__(self, other):
            warnings.warn(f"__rrshift__ is depreciated on ByteTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) >> primitive_conversion(self)

        def __rand__(self, other):
            warnings.warn(f"__rand__ is depreciated on ByteTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) & primitive_conversion(self)

        def __rxor__(self, other):
            warnings.warn(f"__rxor__ is depreciated on ByteTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) ^ primitive_conversion(self)

        def __ror__(self, other):
            warnings.warn(f"__ror__ is depreciated on ByteTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) | primitive_conversion(self)

        def __neg__(self):
            warnings.warn(f"__neg__ is depreciated on ByteTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return -self.value_

        def __pos__(self):
            warnings.warn(f"__pos__ is depreciated on ByteTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return +self.value_

        def __abs__(self):
            warnings.warn(f"__abs__ is depreciated on ByteTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return abs(self.value_)

        def __invert__(self):
            warnings.warn(f"__invert__ is depreciated on ByteTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return ~self.value_

    cdef signed char _sanitise_value(ByteTag self, value):
        return (value & 0x7F) - (value & 0x80)

    cdef str _to_snbt(ByteTag self):
        return f"{self.value_}b"

    cdef void write_payload(
        ByteTag self,
        object buffer: BytesIO,
        bint little_endian,
        string_encoder: EncoderType,
    ) except *:
        write_byte(self.value_, buffer)


cdef inline ShortTag read_short_tag(BufferContext buffer, bint little_endian):
    cdef ShortTag tag = ShortTag.__new__(ShortTag)
    cdef short *pointer = <short*> read_data(buffer, 2)
    tag.value_ = pointer[0]
    to_little_endian(&tag.value_, 2, little_endian)
    return tag


cdef class ShortTag(AbstractBaseIntTag):
    """
    A 2 byte integer class.
    Can Store numbers between -(2^15) and (2^15 - 1)
    """
    tag_id = ID_SHORT

    def __init__(ShortTag self, value = 0):
        self.value_ = self._sanitise_value(int(value))

    def __str__(ShortTag self):
        return str(self.value_)

    def __eq__(ShortTag self, other):
        cdef ShortTag other_
        if isinstance(other, ShortTag):
            other_ = other
            return self.value_ == other_.value_
        elif __major__ <= 2:
            warnings.warn("NBT comparison operator (a == b) will only return True between classes of the same type.", FutureWarning)
            return self.value_ == primitive_conversion(other)
        return NotImplemented

    def __reduce__(ShortTag self):
        return self.__class__, (self.value_,)

    def __deepcopy__(ShortTag self, memo=None):
        return self.__class__(deepcopy(self.value_, memo=memo))

    def __copy__(ShortTag self):
        return self.__class__(self.value_)

    def __hash__(ShortTag self):
        return hash((self.tag_id, self.value_))

    def __ge__(ShortTag self, other):
        cdef ShortTag other_
        if isinstance(other, ShortTag):
            other_ = other
            return self.value_ >= other_.value_
        elif __major__ <= 2:
            warnings.warn("NBT comparison operator (a >= b) will only return True between classes of the same type.", FutureWarning)
            return self.value_ >= primitive_conversion(other)
        return NotImplemented

    def __gt__(ShortTag self, other):
        cdef ShortTag other_
        if isinstance(other, ShortTag):
            other_ = other
            return self.value_ > other_.value_
        elif __major__ <= 2:
            warnings.warn("NBT comparison operator (a > b) will only return True between classes of the same type.", FutureWarning)
            return self.value_ > primitive_conversion(other)
        return NotImplemented

    def __le__(ShortTag self, other):
        cdef ShortTag other_
        if isinstance(other, ShortTag):
            other_ = other
            return self.value_ <= other_.value_
        elif __major__ <= 2:
            warnings.warn("NBT comparison operator (a <= b) will only return True between classes of the same type.", FutureWarning)
            return self.value_ <= primitive_conversion(other)
        return NotImplemented

    def __lt__(ShortTag self, other):
        cdef ShortTag other_
        if isinstance(other, ShortTag):
            other_ = other
            return self.value_ < other_.value_
        elif __major__ <= 2:
            warnings.warn("NBT comparison operator (a == b) will only return True between classes of the same type.", FutureWarning)
            return self.value_ < primitive_conversion(other)
        return NotImplemented

    def __repr__(ShortTag self):
        return f"{self.__class__.__name__}({self.value_})"

    def __int__(ShortTag self):
        return self.value_.__int__()

    def __float__(ShortTag self):
        return self.value_.__float__()

    def __bool__(ShortTag self):
        return self.value_.__bool__()

    @property
    def py_int(ShortTag self) -> int:
        """
        A python int representation of the class.
        The returned data is immutable so changes will not mirror the instance.
        """
        return self.value_

    if __major__ <= 2:
        def __add__(self, other):
            warnings.warn(f"__add__ is depreciated on ShortTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) + primitive_conversion(other)

        def __sub__(self, other):
            warnings.warn(f"__sub__ is depreciated on ShortTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) - primitive_conversion(other)

        def __mul__(self, other):
            warnings.warn(f"__mul__ is depreciated on ShortTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) * primitive_conversion(other)

        def __truediv__(self, other):
            warnings.warn(f"__truediv__ is depreciated on ShortTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) / primitive_conversion(other)

        def __floordiv__(self, other):
            warnings.warn(f"__floordiv__ is depreciated on ShortTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) // primitive_conversion(other)

        def __mod__(self, other):
            warnings.warn(f"__mod__ is depreciated on ShortTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) % primitive_conversion(other)

        def __divmod__(self, other):
            warnings.warn(f"__divmod__ is depreciated on ShortTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return divmod(primitive_conversion(self), primitive_conversion(other))

        def __pow__(self, power, modulo):
            warnings.warn(f"__pow__ is depreciated on ShortTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return pow(primitive_conversion(self), power, modulo)

        def __lshift__(self, other):
            warnings.warn(f"__lshift__ is depreciated on ShortTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) << primitive_conversion(other)

        def __rshift__(self, other):
            warnings.warn(f"__rshift__ is depreciated on ShortTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) >> primitive_conversion(other)

        def __and__(self, other):
            warnings.warn(f"__and__ is depreciated on ShortTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) & primitive_conversion(other)

        def __xor__(self, other):
            warnings.warn(f"__xor__ is depreciated on ShortTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) ^ primitive_conversion(other)

        def __or__(self, other):
            warnings.warn(f"__or__ is depreciated on ShortTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) | primitive_conversion(other)

        def __radd__(self, other):
            warnings.warn(f"__radd__ is depreciated on ShortTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) + primitive_conversion(self)

        def __rsub__(self, other):
            warnings.warn(f"__rsub__ is depreciated on ShortTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) - primitive_conversion(self)

        def __rmul__(self, other):
            warnings.warn(f"__rmul__ is depreciated on ShortTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) * primitive_conversion(self)

        def __rtruediv__(self, other):
            warnings.warn(f"__rtruediv__ is depreciated on ShortTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) / primitive_conversion(self)

        def __rfloordiv__(self, other):
            warnings.warn(f"__rfloordiv__ is depreciated on ShortTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) // primitive_conversion(self)

        def __rmod__(self, other):
            warnings.warn(f"__rmod__ is depreciated on ShortTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) % primitive_conversion(self)

        def __rdivmod__(self, other):
            warnings.warn(f"__rdivmod__ is depreciated on ShortTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return divmod(primitive_conversion(other), primitive_conversion(self))

        def __rpow__(self, other, modulo):
            warnings.warn(f"__rpow__ is depreciated on ShortTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return pow(primitive_conversion(other), primitive_conversion(self), modulo)

        def __rlshift__(self, other):
            warnings.warn(f"__rlshift__ is depreciated on ShortTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) << primitive_conversion(self)

        def __rrshift__(self, other):
            warnings.warn(f"__rrshift__ is depreciated on ShortTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) >> primitive_conversion(self)

        def __rand__(self, other):
            warnings.warn(f"__rand__ is depreciated on ShortTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) & primitive_conversion(self)

        def __rxor__(self, other):
            warnings.warn(f"__rxor__ is depreciated on ShortTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) ^ primitive_conversion(self)

        def __ror__(self, other):
            warnings.warn(f"__ror__ is depreciated on ShortTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) | primitive_conversion(self)

        def __neg__(self):
            warnings.warn(f"__neg__ is depreciated on ShortTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return -self.value_

        def __pos__(self):
            warnings.warn(f"__pos__ is depreciated on ShortTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return +self.value_

        def __abs__(self):
            warnings.warn(f"__abs__ is depreciated on ShortTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return abs(self.value_)

        def __invert__(self):
            warnings.warn(f"__invert__ is depreciated on ShortTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return ~self.value_

    cdef short _sanitise_value(ShortTag self, value):
        return (value & 0x7FFF) - (value & 0x8000)

    cdef str _to_snbt(ShortTag self):
        return f"{self.value_}s"

    cdef void write_payload(
        ShortTag self,
        object buffer: BytesIO,
        bint little_endian,
        string_encoder: EncoderType,
    ) except *:
        write_short(self.value_, buffer, little_endian)


cdef inline IntTag read_int_tag(BufferContext buffer, bint little_endian):
    cdef IntTag tag = IntTag.__new__(IntTag)
    cdef int*pointer = <int*> read_data(buffer, 4)
    tag.value_ = pointer[0]
    to_little_endian(&tag.value_, 4, little_endian)
    return tag


cdef class IntTag(AbstractBaseIntTag):
    """
    A 4 byte integer class.
    Can Store numbers between -(2^31) and (2^31 - 1)
    """
    tag_id = ID_INT

    def __init__(IntTag self, value = 0):
        self.value_ = self._sanitise_value(int(value))

    def __str__(IntTag self):
        return str(self.value_)

    def __eq__(IntTag self, other):
        cdef IntTag other_
        if isinstance(other, IntTag):
            other_ = other
            return self.value_ == other_.value_
        elif __major__ <= 2:
            warnings.warn("NBT comparison operator (a == b) will only return True between classes of the same type.", FutureWarning)
            return self.value_ == primitive_conversion(other)
        return NotImplemented

    def __reduce__(IntTag self):
        return self.__class__, (self.value_,)

    def __deepcopy__(IntTag self, memo=None):
        return self.__class__(deepcopy(self.value_, memo=memo))

    def __copy__(IntTag self):
        return self.__class__(self.value_)

    def __hash__(IntTag self):
        return hash((self.tag_id, self.value_))

    def __ge__(IntTag self, other):
        cdef IntTag other_
        if isinstance(other, IntTag):
            other_ = other
            return self.value_ >= other_.value_
        elif __major__ <= 2:
            warnings.warn("NBT comparison operator (a >= b) will only return True between classes of the same type.", FutureWarning)
            return self.value_ >= primitive_conversion(other)
        return NotImplemented

    def __gt__(IntTag self, other):
        cdef IntTag other_
        if isinstance(other, IntTag):
            other_ = other
            return self.value_ > other_.value_
        elif __major__ <= 2:
            warnings.warn("NBT comparison operator (a > b) will only return True between classes of the same type.", FutureWarning)
            return self.value_ > primitive_conversion(other)
        return NotImplemented

    def __le__(IntTag self, other):
        cdef IntTag other_
        if isinstance(other, IntTag):
            other_ = other
            return self.value_ <= other_.value_
        elif __major__ <= 2:
            warnings.warn("NBT comparison operator (a <= b) will only return True between classes of the same type.", FutureWarning)
            return self.value_ <= primitive_conversion(other)
        return NotImplemented

    def __lt__(IntTag self, other):
        cdef IntTag other_
        if isinstance(other, IntTag):
            other_ = other
            return self.value_ < other_.value_
        elif __major__ <= 2:
            warnings.warn("NBT comparison operator (a == b) will only return True between classes of the same type.", FutureWarning)
            return self.value_ < primitive_conversion(other)
        return NotImplemented

    def __repr__(IntTag self):
        return f"{self.__class__.__name__}({self.value_})"

    def __int__(IntTag self):
        return self.value_.__int__()

    def __float__(IntTag self):
        return self.value_.__float__()

    def __bool__(IntTag self):
        return self.value_.__bool__()

    @property
    def py_int(IntTag self) -> int:
        """
        A python int representation of the class.
        The returned data is immutable so changes will not mirror the instance.
        """
        return self.value_

    if __major__ <= 2:
        def __add__(self, other):
            warnings.warn(f"__add__ is depreciated on IntTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) + primitive_conversion(other)

        def __sub__(self, other):
            warnings.warn(f"__sub__ is depreciated on IntTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) - primitive_conversion(other)

        def __mul__(self, other):
            warnings.warn(f"__mul__ is depreciated on IntTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) * primitive_conversion(other)

        def __truediv__(self, other):
            warnings.warn(f"__truediv__ is depreciated on IntTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) / primitive_conversion(other)

        def __floordiv__(self, other):
            warnings.warn(f"__floordiv__ is depreciated on IntTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) // primitive_conversion(other)

        def __mod__(self, other):
            warnings.warn(f"__mod__ is depreciated on IntTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) % primitive_conversion(other)

        def __divmod__(self, other):
            warnings.warn(f"__divmod__ is depreciated on IntTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return divmod(primitive_conversion(self), primitive_conversion(other))

        def __pow__(self, power, modulo):
            warnings.warn(f"__pow__ is depreciated on IntTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return pow(primitive_conversion(self), power, modulo)

        def __lshift__(self, other):
            warnings.warn(f"__lshift__ is depreciated on IntTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) << primitive_conversion(other)

        def __rshift__(self, other):
            warnings.warn(f"__rshift__ is depreciated on IntTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) >> primitive_conversion(other)

        def __and__(self, other):
            warnings.warn(f"__and__ is depreciated on IntTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) & primitive_conversion(other)

        def __xor__(self, other):
            warnings.warn(f"__xor__ is depreciated on IntTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) ^ primitive_conversion(other)

        def __or__(self, other):
            warnings.warn(f"__or__ is depreciated on IntTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) | primitive_conversion(other)

        def __radd__(self, other):
            warnings.warn(f"__radd__ is depreciated on IntTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) + primitive_conversion(self)

        def __rsub__(self, other):
            warnings.warn(f"__rsub__ is depreciated on IntTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) - primitive_conversion(self)

        def __rmul__(self, other):
            warnings.warn(f"__rmul__ is depreciated on IntTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) * primitive_conversion(self)

        def __rtruediv__(self, other):
            warnings.warn(f"__rtruediv__ is depreciated on IntTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) / primitive_conversion(self)

        def __rfloordiv__(self, other):
            warnings.warn(f"__rfloordiv__ is depreciated on IntTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) // primitive_conversion(self)

        def __rmod__(self, other):
            warnings.warn(f"__rmod__ is depreciated on IntTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) % primitive_conversion(self)

        def __rdivmod__(self, other):
            warnings.warn(f"__rdivmod__ is depreciated on IntTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return divmod(primitive_conversion(other), primitive_conversion(self))

        def __rpow__(self, other, modulo):
            warnings.warn(f"__rpow__ is depreciated on IntTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return pow(primitive_conversion(other), primitive_conversion(self), modulo)

        def __rlshift__(self, other):
            warnings.warn(f"__rlshift__ is depreciated on IntTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) << primitive_conversion(self)

        def __rrshift__(self, other):
            warnings.warn(f"__rrshift__ is depreciated on IntTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) >> primitive_conversion(self)

        def __rand__(self, other):
            warnings.warn(f"__rand__ is depreciated on IntTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) & primitive_conversion(self)

        def __rxor__(self, other):
            warnings.warn(f"__rxor__ is depreciated on IntTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) ^ primitive_conversion(self)

        def __ror__(self, other):
            warnings.warn(f"__ror__ is depreciated on IntTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) | primitive_conversion(self)

        def __neg__(self):
            warnings.warn(f"__neg__ is depreciated on IntTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return -self.value_

        def __pos__(self):
            warnings.warn(f"__pos__ is depreciated on IntTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return +self.value_

        def __abs__(self):
            warnings.warn(f"__abs__ is depreciated on IntTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return abs(self.value_)

        def __invert__(self):
            warnings.warn(f"__invert__ is depreciated on IntTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return ~self.value_

    cdef int _sanitise_value(IntTag self, value):
        return (value & 0x7FFF_FFFF) - (value & 0x8000_0000)

    cdef str _to_snbt(IntTag self):
        return f"{self.value_}"

    cdef void write_payload(
        IntTag self,
        object buffer: BytesIO,
        bint little_endian,
        string_encoder: EncoderType,
    ) except *:
        write_int(self.value_, buffer, little_endian)


cdef inline LongTag read_long_tag(BufferContext buffer, bint little_endian):
    cdef LongTag tag = LongTag.__new__(LongTag)
    cdef long long *pointer = <long long *> read_data(buffer, 8)
    tag.value_ = pointer[0]
    to_little_endian(&tag.value_, 8, little_endian)
    return tag


cdef class LongTag(AbstractBaseIntTag):
    """
    An 8 byte integer class.
    Can Store numbers between -(2^63) and (2^63 - 1)
    """
    tag_id = ID_LONG

    def __init__(LongTag self, value = 0):
        self.value_ = self._sanitise_value(int(value))

    def __str__(LongTag self):
        return str(self.value_)

    def __eq__(LongTag self, other):
        cdef LongTag other_
        if isinstance(other, LongTag):
            other_ = other
            return self.value_ == other_.value_
        elif __major__ <= 2:
            warnings.warn("NBT comparison operator (a == b) will only return True between classes of the same type.", FutureWarning)
            return self.value_ == primitive_conversion(other)
        return NotImplemented

    def __reduce__(LongTag self):
        return self.__class__, (self.value_,)

    def __deepcopy__(LongTag self, memo=None):
        return self.__class__(deepcopy(self.value_, memo=memo))

    def __copy__(LongTag self):
        return self.__class__(self.value_)

    def __hash__(LongTag self):
        return hash((self.tag_id, self.value_))

    def __ge__(LongTag self, other):
        cdef LongTag other_
        if isinstance(other, LongTag):
            other_ = other
            return self.value_ >= other_.value_
        elif __major__ <= 2:
            warnings.warn("NBT comparison operator (a >= b) will only return True between classes of the same type.", FutureWarning)
            return self.value_ >= primitive_conversion(other)
        return NotImplemented

    def __gt__(LongTag self, other):
        cdef LongTag other_
        if isinstance(other, LongTag):
            other_ = other
            return self.value_ > other_.value_
        elif __major__ <= 2:
            warnings.warn("NBT comparison operator (a > b) will only return True between classes of the same type.", FutureWarning)
            return self.value_ > primitive_conversion(other)
        return NotImplemented

    def __le__(LongTag self, other):
        cdef LongTag other_
        if isinstance(other, LongTag):
            other_ = other
            return self.value_ <= other_.value_
        elif __major__ <= 2:
            warnings.warn("NBT comparison operator (a <= b) will only return True between classes of the same type.", FutureWarning)
            return self.value_ <= primitive_conversion(other)
        return NotImplemented

    def __lt__(LongTag self, other):
        cdef LongTag other_
        if isinstance(other, LongTag):
            other_ = other
            return self.value_ < other_.value_
        elif __major__ <= 2:
            warnings.warn("NBT comparison operator (a == b) will only return True between classes of the same type.", FutureWarning)
            return self.value_ < primitive_conversion(other)
        return NotImplemented

    def __repr__(LongTag self):
        return f"{self.__class__.__name__}({self.value_})"

    def __int__(LongTag self):
        return self.value_.__int__()

    def __float__(LongTag self):
        return self.value_.__float__()

    def __bool__(LongTag self):
        return self.value_.__bool__()

    @property
    def py_int(LongTag self) -> int:
        """
        A python int representation of the class.
        The returned data is immutable so changes will not mirror the instance.
        """
        return self.value_

    if __major__ <= 2:
        def __add__(self, other):
            warnings.warn(f"__add__ is depreciated on LongTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) + primitive_conversion(other)

        def __sub__(self, other):
            warnings.warn(f"__sub__ is depreciated on LongTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) - primitive_conversion(other)

        def __mul__(self, other):
            warnings.warn(f"__mul__ is depreciated on LongTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) * primitive_conversion(other)

        def __truediv__(self, other):
            warnings.warn(f"__truediv__ is depreciated on LongTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) / primitive_conversion(other)

        def __floordiv__(self, other):
            warnings.warn(f"__floordiv__ is depreciated on LongTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) // primitive_conversion(other)

        def __mod__(self, other):
            warnings.warn(f"__mod__ is depreciated on LongTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) % primitive_conversion(other)

        def __divmod__(self, other):
            warnings.warn(f"__divmod__ is depreciated on LongTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return divmod(primitive_conversion(self), primitive_conversion(other))

        def __pow__(self, power, modulo):
            warnings.warn(f"__pow__ is depreciated on LongTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return pow(primitive_conversion(self), power, modulo)

        def __lshift__(self, other):
            warnings.warn(f"__lshift__ is depreciated on LongTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) << primitive_conversion(other)

        def __rshift__(self, other):
            warnings.warn(f"__rshift__ is depreciated on LongTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) >> primitive_conversion(other)

        def __and__(self, other):
            warnings.warn(f"__and__ is depreciated on LongTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) & primitive_conversion(other)

        def __xor__(self, other):
            warnings.warn(f"__xor__ is depreciated on LongTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) ^ primitive_conversion(other)

        def __or__(self, other):
            warnings.warn(f"__or__ is depreciated on LongTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) | primitive_conversion(other)

        def __radd__(self, other):
            warnings.warn(f"__radd__ is depreciated on LongTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) + primitive_conversion(self)

        def __rsub__(self, other):
            warnings.warn(f"__rsub__ is depreciated on LongTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) - primitive_conversion(self)

        def __rmul__(self, other):
            warnings.warn(f"__rmul__ is depreciated on LongTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) * primitive_conversion(self)

        def __rtruediv__(self, other):
            warnings.warn(f"__rtruediv__ is depreciated on LongTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) / primitive_conversion(self)

        def __rfloordiv__(self, other):
            warnings.warn(f"__rfloordiv__ is depreciated on LongTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) // primitive_conversion(self)

        def __rmod__(self, other):
            warnings.warn(f"__rmod__ is depreciated on LongTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) % primitive_conversion(self)

        def __rdivmod__(self, other):
            warnings.warn(f"__rdivmod__ is depreciated on LongTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return divmod(primitive_conversion(other), primitive_conversion(self))

        def __rpow__(self, other, modulo):
            warnings.warn(f"__rpow__ is depreciated on LongTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return pow(primitive_conversion(other), primitive_conversion(self), modulo)

        def __rlshift__(self, other):
            warnings.warn(f"__rlshift__ is depreciated on LongTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) << primitive_conversion(self)

        def __rrshift__(self, other):
            warnings.warn(f"__rrshift__ is depreciated on LongTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) >> primitive_conversion(self)

        def __rand__(self, other):
            warnings.warn(f"__rand__ is depreciated on LongTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) & primitive_conversion(self)

        def __rxor__(self, other):
            warnings.warn(f"__rxor__ is depreciated on LongTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) ^ primitive_conversion(self)

        def __ror__(self, other):
            warnings.warn(f"__ror__ is depreciated on LongTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) | primitive_conversion(self)

        def __neg__(self):
            warnings.warn(f"__neg__ is depreciated on LongTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return -self.value_

        def __pos__(self):
            warnings.warn(f"__pos__ is depreciated on LongTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return +self.value_

        def __abs__(self):
            warnings.warn(f"__abs__ is depreciated on LongTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return abs(self.value_)

        def __invert__(self):
            warnings.warn(f"__invert__ is depreciated on LongTag and will be removed in the future. Please use .py_int to achieve the same behaviour.", DeprecationWarning)
            return ~self.value_

    cdef long long _sanitise_value(LongTag self, value):
        return (value & 0x7FFF_FFFF_FFFF_FFFF) - (value & 0x8000_0000_0000_0000)

    cdef str _to_snbt(LongTag self):
        return f"{self.value_}L"

    cdef void write_payload(
        LongTag self,
        object buffer: BytesIO,
        bint little_endian,
        string_encoder: EncoderType,
    ) except *:
        write_long(self.value_, buffer, little_endian)



## This file is generated by tempita. Do not modify this file directly or your changes will get overwritten.
## To edit this file edit the template in template/src

from io import BytesIO
from copy import deepcopy
import warnings

from . import __major__
from ._value cimport AbstractBaseImmutableTag
from ._const cimport ID_STRING
from ._util cimport write_string, read_string, BufferContext
if __major__ <= 2:
    from ._util import primitive_conversion
from ._dtype import EncoderType


cdef inline escape(str string):
    return string.replace('\\', '\\\\').replace('"', '\\"')


cdef inline StringTag read_string_tag(BufferContext buffer, bint little_endian, string_decoder: DecoderType):
    cdef StringTag tag = StringTag.__new__(StringTag)
    tag.value_ = read_string(buffer, little_endian, string_decoder)
    return tag


cdef class StringTag(AbstractBaseImmutableTag):
    """A class that behaves like a string."""
    tag_id = ID_STRING

    def __init__(StringTag self, value = ""):
        if isinstance(value, bytes):
            warnings.warn("StringTag no longer takes bytes as input. You will need to decode this to a string before giving it here. The string of the object will be used instead.", DeprecationWarning)
        self.value_ = str(value)

    def __str__(StringTag self):
        return str(self.value_)

    def __eq__(StringTag self, other):
        cdef StringTag other_
        if isinstance(other, StringTag):
            other_ = other
            return self.value_ == other_.value_
        elif __major__ <= 2:
            warnings.warn("NBT comparison operator (a == b) will only return True between classes of the same type.", FutureWarning)
            return self.value_ == primitive_conversion(other)
        return NotImplemented

    def __reduce__(StringTag self):
        return self.__class__, (self.value_,)

    def __deepcopy__(StringTag self, memo=None):
        return self.__class__(deepcopy(self.value_, memo=memo))

    def __copy__(StringTag self):
        return self.__class__(self.value_)

    def __hash__(StringTag self):
        return hash((self.tag_id, self.value_))

    def __ge__(StringTag self, other):
        cdef StringTag other_
        if isinstance(other, StringTag):
            other_ = other
            return self.value_ >= other_.value_
        elif __major__ <= 2:
            warnings.warn("NBT comparison operator (a >= b) will only return True between classes of the same type.", FutureWarning)
            return self.value_ >= primitive_conversion(other)
        return NotImplemented

    def __gt__(StringTag self, other):
        cdef StringTag other_
        if isinstance(other, StringTag):
            other_ = other
            return self.value_ > other_.value_
        elif __major__ <= 2:
            warnings.warn("NBT comparison operator (a > b) will only return True between classes of the same type.", FutureWarning)
            return self.value_ > primitive_conversion(other)
        return NotImplemented

    def __le__(StringTag self, other):
        cdef StringTag other_
        if isinstance(other, StringTag):
            other_ = other
            return self.value_ <= other_.value_
        elif __major__ <= 2:
            warnings.warn("NBT comparison operator (a <= b) will only return True between classes of the same type.", FutureWarning)
            return self.value_ <= primitive_conversion(other)
        return NotImplemented

    def __lt__(StringTag self, other):
        cdef StringTag other_
        if isinstance(other, StringTag):
            other_ = other
            return self.value_ < other_.value_
        elif __major__ <= 2:
            warnings.warn("NBT comparison operator (a == b) will only return True between classes of the same type.", FutureWarning)
            return self.value_ < primitive_conversion(other)
        return NotImplemented

    @property
    def py_str(StringTag self) -> str:
        """
        A python string representation of the class.
        """
        return self.value_

    @property
    def py_data(self):
        """
        A python representation of the class. Note that the return type is undefined and may change in the future.
        You would be better off using the py_{type} or np_array properties if you require a fixed type.
        This is here for convenience to get a python representation under the same property name.
        """
        return self.py_str

    def __repr__(StringTag self):
        return f"{self.__class__.__name__}(\"{escape(self.py_str)}\")"

    cdef str _to_snbt(StringTag self):
        return f"\"{escape(self.py_str)}\""

    cdef void write_payload(
        StringTag self,
        object buffer: BytesIO,
        bint little_endian,
        string_encoder: EncoderType,
    ) except *:
        write_string(self.value_, buffer, little_endian, string_encoder)

    def __len__(self) -> int:
        return len(self.value_)

    if __major__ <= 2:
        def __getitem__(self, item) -> str:
            warnings.warn(f"__getitem__ is depreciated on StringTag and will be removed in the future. Please use .py_str to achieve the same behaviour.", DeprecationWarning)
            return self.value_[item]

        def __add__(self, other):
            warnings.warn(f"__add__ is depreciated on StringTag and will be removed in the future. Please use .py_str to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) + primitive_conversion(other)

        def __radd__(self, other):
            warnings.warn(f"__radd__ is depreciated on StringTag and will be removed in the future. Please use .py_str to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) + primitive_conversion(self)

        def __mul__(self, other):
            warnings.warn(f"__mul__ is depreciated on StringTag and will be removed in the future. Please use .py_str to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) * primitive_conversion(other)

        def __rmul__(self, other):
            warnings.warn(f"__rmul__ is depreciated on StringTag and will be removed in the future. Please use .py_str to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) * primitive_conversion(self)



from ._value cimport AbstractBaseMutableTag
from ._util cimport BufferContext
from ._dtype import DecoderType
from ._int cimport (
    ByteTag,
    ShortTag,
    IntTag,
    LongTag,
)
from ._float cimport (
    FloatTag,
    DoubleTag,
)
from ._array cimport (
    ByteArrayTag,
    IntArrayTag,
    LongArrayTag,
)
from ._string cimport StringTag
from ._list cimport CyListTag


cdef CyCompoundTag read_compound_tag(BufferContext buffer, bint little_endian, string_decoder: DecoderType)

cdef class CyCompoundTag(AbstractBaseMutableTag):
    cdef dict value_
    
    cpdef ByteTag get_byte(self, str key, ByteTag default=*)
    cpdef ShortTag get_short(self, str key, ShortTag default=*)
    cpdef IntTag get_int(self, str key, IntTag default=*)
    cpdef LongTag get_long(self, str key, LongTag default=*)
    cpdef FloatTag get_float(self, str key, FloatTag default=*)
    cpdef DoubleTag get_double(self, str key, DoubleTag default=*)
    cpdef StringTag get_string(self, str key, StringTag default=*)
    cpdef CyListTag get_list(self, str key, CyListTag default=*)
    cpdef CyCompoundTag get_compound(self, str key, CyCompoundTag default=*)
    cpdef ByteArrayTag get_byte_array(self, str key, ByteArrayTag default=*)
    cpdef IntArrayTag get_int_array(self, str key, IntArrayTag default=*)
    cpdef LongArrayTag get_long_array(self, str key, LongArrayTag default=*)

    cpdef ByteTag setdefault_byte(self, str key, ByteTag default= *)
    cpdef ShortTag setdefault_short(self, str key, ShortTag default= *)
    cpdef IntTag setdefault_int(self, str key, IntTag default= *)
    cpdef LongTag setdefault_long(self, str key, LongTag default= *)
    cpdef FloatTag setdefault_float(self, str key, FloatTag default= *)
    cpdef DoubleTag setdefault_double(self, str key, DoubleTag default= *)
    cpdef StringTag setdefault_string(self, str key, StringTag default= *)
    cpdef CyListTag setdefault_list(self, str key, CyListTag default= *)
    cpdef CyCompoundTag setdefault_compound(self, str key, CyCompoundTag default= *)
    cpdef ByteArrayTag setdefault_byte_array(self, str key, ByteArrayTag default= *)
    cpdef IntArrayTag setdefault_int_array(self, str key, IntArrayTag default= *)
    cpdef LongArrayTag setdefault_long_array(self, str key, LongArrayTag default= *)



from ._numeric cimport AbstractBaseNumericTag
from ._util cimport BufferContext

cdef class AbstractBaseIntTag(AbstractBaseNumericTag):
    pass

cdef ByteTag read_byte_tag(BufferContext buffer, bint little_endian)

cdef class ByteTag(AbstractBaseIntTag):
    cdef signed char value_
    cdef signed char _sanitise_value(self, value)

cdef ShortTag read_short_tag(BufferContext buffer, bint little_endian)

cdef class ShortTag(AbstractBaseIntTag):
    cdef short value_
    cdef short _sanitise_value(self, value)

cdef IntTag read_int_tag(BufferContext buffer, bint little_endian)

cdef class IntTag(AbstractBaseIntTag):
    cdef int value_
    cdef int _sanitise_value(self, value)

cdef LongTag read_long_tag(BufferContext buffer, bint little_endian)

cdef class LongTag(AbstractBaseIntTag):
    cdef long long value_
    cdef long long _sanitise_value(self, value)



from ._dtype import EncoderType


cdef class AbstractBase:
    cpdef str to_snbt(
        self,
        object indent=*,
        object indent_chr=*
    )


cdef class AbstractBaseTag(AbstractBase):
    cdef str _to_snbt(self)

    cdef str _pretty_to_snbt(
        self,
        str indent_chr,
        int indent_count=*,
        bint leading_indent=*
    )

    cdef void write_tag(
        self,
        object buffer,
        str name,
        bint little_endian,
        string_encoder: EncoderType,
    ) except *

    cdef void write_payload(
        self,
        object buffer,
        bint little_endian,
        string_encoder: EncoderType,
    ) except *

cdef class AbstractBaseImmutableTag(AbstractBaseTag):
    pass

cdef class AbstractBaseMutableTag(AbstractBaseTag):
    pass



## This file is generated by tempita. Do not modify this file directly or your changes will get overwritten.
## To edit this file edit the template in template/src

from io import BytesIO
from copy import deepcopy
import warnings
from math import trunc, floor, ceil

from . import __major__
from ._numeric cimport AbstractBaseNumericTag
from ._const cimport ID_FLOAT, ID_DOUBLE
from ._util cimport write_float, write_double, BufferContext, read_data, to_little_endian
if __major__ <= 2:
    from ._util import primitive_conversion
from ._dtype import EncoderType


cdef class AbstractBaseFloatTag(AbstractBaseNumericTag):
    """Abstract Base Class for all float Tag classes"""

    @property
    def py_float(AbstractBaseNumericTag self) -> float:
        """
        A python float representation of the class.
        The returned data is immutable so changes will not mirror the instance.
        """
        raise NotImplementedError

    @property
    def py_data(self):
        """
        A python representation of the class. Note that the return type is undefined and may change in the future.
        You would be better off using the py_{type} or np_array properties if you require a fixed type.
        This is here for convenience to get a python representation under the same property name.
        """
        return self.py_float



cdef inline FloatTag read_float_tag(BufferContext buffer, bint little_endian):
    cdef FloatTag tag = FloatTag.__new__(FloatTag)
    cdef float*pointer = <float*> read_data(buffer, 4)
    tag.value_ = pointer[0]
    to_little_endian(&tag.value_, 4, little_endian)
    return tag


cdef class FloatTag(AbstractBaseFloatTag):
    """A single precision float class."""
    tag_id = ID_FLOAT

    def __init__(FloatTag self, value = 0):
        self.value_ = float(value)

    def __str__(FloatTag self):
        return str(self.value_)

    def __eq__(FloatTag self, other):
        cdef FloatTag other_
        if isinstance(other, FloatTag):
            other_ = other
            return self.value_ == other_.value_
        elif __major__ <= 2:
            warnings.warn("NBT comparison operator (a == b) will only return True between classes of the same type.", FutureWarning)
            return self.value_ == primitive_conversion(other)
        return NotImplemented

    def __reduce__(FloatTag self):
        return self.__class__, (self.value_,)

    def __deepcopy__(FloatTag self, memo=None):
        return self.__class__(deepcopy(self.value_, memo=memo))

    def __copy__(FloatTag self):
        return self.__class__(self.value_)

    def __hash__(FloatTag self):
        return hash((self.tag_id, self.value_))

    def __ge__(FloatTag self, other):
        cdef FloatTag other_
        if isinstance(other, FloatTag):
            other_ = other
            return self.value_ >= other_.value_
        elif __major__ <= 2:
            warnings.warn("NBT comparison operator (a >= b) will only return True between classes of the same type.", FutureWarning)
            return self.value_ >= primitive_conversion(other)
        return NotImplemented

    def __gt__(FloatTag self, other):
        cdef FloatTag other_
        if isinstance(other, FloatTag):
            other_ = other
            return self.value_ > other_.value_
        elif __major__ <= 2:
            warnings.warn("NBT comparison operator (a > b) will only return True between classes of the same type.", FutureWarning)
            return self.value_ > primitive_conversion(other)
        return NotImplemented

    def __le__(FloatTag self, other):
        cdef FloatTag other_
        if isinstance(other, FloatTag):
            other_ = other
            return self.value_ <= other_.value_
        elif __major__ <= 2:
            warnings.warn("NBT comparison operator (a <= b) will only return True between classes of the same type.", FutureWarning)
            return self.value_ <= primitive_conversion(other)
        return NotImplemented

    def __lt__(FloatTag self, other):
        cdef FloatTag other_
        if isinstance(other, FloatTag):
            other_ = other
            return self.value_ < other_.value_
        elif __major__ <= 2:
            warnings.warn("NBT comparison operator (a == b) will only return True between classes of the same type.", FutureWarning)
            return self.value_ < primitive_conversion(other)
        return NotImplemented

    def __repr__(FloatTag self):
        return f"{self.__class__.__name__}({self.value_})"

    def __int__(FloatTag self):
        return self.value_.__int__()

    def __float__(FloatTag self):
        return self.value_.__float__()

    def __bool__(FloatTag self):
        return self.value_.__bool__()

    @property
    def py_float(FloatTag self) -> float:
        """
        A python float representation of the class.
        The returned data is immutable so changes will not mirror the instance.
        """
        return self.value_

    if __major__ <= 2:
        def __add__(self, other):
            warnings.warn(f"__add__ is depreciated on FloatTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return float(primitive_conversion(self) + primitive_conversion(other))

        def __sub__(self, other):
            warnings.warn(f"__sub__ is depreciated on FloatTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return float(primitive_conversion(self) - primitive_conversion(other))

        def __mul__(self, other):
            warnings.warn(f"__mul__ is depreciated on FloatTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return float(primitive_conversion(self) * primitive_conversion(other))

        def __truediv__(self, other):
            warnings.warn(f"__truediv__ is depreciated on FloatTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return float(primitive_conversion(self) / primitive_conversion(other))

        def __floordiv__(self, other):
            warnings.warn(f"__floordiv__ is depreciated on FloatTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) // primitive_conversion(other)

        def __mod__(self, other):
            warnings.warn(f"__mod__ is depreciated on FloatTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) % primitive_conversion(other)

        def __divmod__(self, other):
            warnings.warn(f"__divmod__ is depreciated on FloatTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return divmod(primitive_conversion(self), primitive_conversion(other))

        def __pow__(self, power, modulo):
            warnings.warn(f"__pow__ is depreciated on FloatTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return pow(primitive_conversion(self), power, modulo)

        def __lshift__(self, other):
            warnings.warn(f"__lshift__ is depreciated on FloatTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) << primitive_conversion(other)

        def __rshift__(self, other):
            warnings.warn(f"__rshift__ is depreciated on FloatTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) >> primitive_conversion(other)

        def __and__(self, other):
            warnings.warn(f"__and__ is depreciated on FloatTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) & primitive_conversion(other)

        def __xor__(self, other):
            warnings.warn(f"__xor__ is depreciated on FloatTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) ^ primitive_conversion(other)

        def __or__(self, other):
            warnings.warn(f"__or__ is depreciated on FloatTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) | primitive_conversion(other)

        def __radd__(self, other):
            warnings.warn(f"__radd__ is depreciated on FloatTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return float(primitive_conversion(other) + primitive_conversion(self))

        def __rsub__(self, other):
            warnings.warn(f"__rsub__ is depreciated on FloatTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return float(primitive_conversion(other) - primitive_conversion(self))

        def __rmul__(self, other):
            warnings.warn(f"__rmul__ is depreciated on FloatTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) * primitive_conversion(self)

        def __rtruediv__(self, other):
            warnings.warn(f"__rtruediv__ is depreciated on FloatTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return float(primitive_conversion(other) / primitive_conversion(self))

        def __rfloordiv__(self, other):
            warnings.warn(f"__rfloordiv__ is depreciated on FloatTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) // primitive_conversion(self)

        def __rmod__(self, other):
            warnings.warn(f"__rmod__ is depreciated on FloatTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) % primitive_conversion(self)

        def __rdivmod__(self, other):
            warnings.warn(f"__rdivmod__ is depreciated on FloatTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return divmod(primitive_conversion(other), primitive_conversion(self))

        def __rpow__(self, other, modulo):
            warnings.warn(f"__rpow__ is depreciated on FloatTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return pow(primitive_conversion(other), primitive_conversion(self), modulo)

        def __rlshift__(self, other):
            warnings.warn(f"__rlshift__ is depreciated on FloatTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) << primitive_conversion(self)

        def __rrshift__(self, other):
            warnings.warn(f"__rrshift__ is depreciated on FloatTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) >> primitive_conversion(self)

        def __rand__(self, other):
            warnings.warn(f"__rand__ is depreciated on FloatTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) & primitive_conversion(self)

        def __rxor__(self, other):
            warnings.warn(f"__rxor__ is depreciated on FloatTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) ^ primitive_conversion(self)

        def __ror__(self, other):
            warnings.warn(f"__ror__ is depreciated on FloatTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) | primitive_conversion(self)

        def __neg__(self):
            warnings.warn(f"__neg__ is depreciated on FloatTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return -self.value_

        def __pos__(self):
            warnings.warn(f"__pos__ is depreciated on FloatTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return +self.value_

        def __abs__(self):
            warnings.warn(f"__abs__ is depreciated on FloatTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return abs(self.value_)

        def __round__(self, n=None):
            warnings.warn(f"__round__ is depreciated on FloatTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return round(self.value_, n)

        def __trunc__(self):
            warnings.warn(f"__trunc__ is depreciated on FloatTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return trunc(self.value_)

        def __floor__(self):
            warnings.warn(f"__floor__ is depreciated on FloatTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return floor(self.value_)

        def __ceil__(self):
            warnings.warn(f"__ceil__ is depreciated on FloatTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return ceil(self.value_)

    cdef str _to_snbt(FloatTag self):
        return f"{self.value_}f"

    cdef void write_payload(
        FloatTag self,
        object buffer: BytesIO,
        bint little_endian,
        string_encoder: EncoderType,
    ) except *:
        write_float(self.value_, buffer, little_endian)


cdef inline DoubleTag read_double_tag(BufferContext buffer, bint little_endian):
    cdef DoubleTag tag = DoubleTag.__new__(DoubleTag)
    cdef double *pointer = <double *> read_data(buffer, 8)
    tag.value_ = pointer[0]
    to_little_endian(&tag.value_, 8, little_endian)
    return tag


cdef class DoubleTag(AbstractBaseFloatTag):
    """A double precision float class."""
    tag_id = ID_DOUBLE

    def __init__(DoubleTag self, value = 0):
        self.value_ = float(value)

    def __str__(DoubleTag self):
        return str(self.value_)

    def __eq__(DoubleTag self, other):
        cdef DoubleTag other_
        if isinstance(other, DoubleTag):
            other_ = other
            return self.value_ == other_.value_
        elif __major__ <= 2:
            warnings.warn("NBT comparison operator (a == b) will only return True between classes of the same type.", FutureWarning)
            return self.value_ == primitive_conversion(other)
        return NotImplemented

    def __reduce__(DoubleTag self):
        return self.__class__, (self.value_,)

    def __deepcopy__(DoubleTag self, memo=None):
        return self.__class__(deepcopy(self.value_, memo=memo))

    def __copy__(DoubleTag self):
        return self.__class__(self.value_)

    def __hash__(DoubleTag self):
        return hash((self.tag_id, self.value_))

    def __ge__(DoubleTag self, other):
        cdef DoubleTag other_
        if isinstance(other, DoubleTag):
            other_ = other
            return self.value_ >= other_.value_
        elif __major__ <= 2:
            warnings.warn("NBT comparison operator (a >= b) will only return True between classes of the same type.", FutureWarning)
            return self.value_ >= primitive_conversion(other)
        return NotImplemented

    def __gt__(DoubleTag self, other):
        cdef DoubleTag other_
        if isinstance(other, DoubleTag):
            other_ = other
            return self.value_ > other_.value_
        elif __major__ <= 2:
            warnings.warn("NBT comparison operator (a > b) will only return True between classes of the same type.", FutureWarning)
            return self.value_ > primitive_conversion(other)
        return NotImplemented

    def __le__(DoubleTag self, other):
        cdef DoubleTag other_
        if isinstance(other, DoubleTag):
            other_ = other
            return self.value_ <= other_.value_
        elif __major__ <= 2:
            warnings.warn("NBT comparison operator (a <= b) will only return True between classes of the same type.", FutureWarning)
            return self.value_ <= primitive_conversion(other)
        return NotImplemented

    def __lt__(DoubleTag self, other):
        cdef DoubleTag other_
        if isinstance(other, DoubleTag):
            other_ = other
            return self.value_ < other_.value_
        elif __major__ <= 2:
            warnings.warn("NBT comparison operator (a == b) will only return True between classes of the same type.", FutureWarning)
            return self.value_ < primitive_conversion(other)
        return NotImplemented

    def __repr__(DoubleTag self):
        return f"{self.__class__.__name__}({self.value_})"

    def __int__(DoubleTag self):
        return self.value_.__int__()

    def __float__(DoubleTag self):
        return self.value_.__float__()

    def __bool__(DoubleTag self):
        return self.value_.__bool__()

    @property
    def py_float(DoubleTag self) -> float:
        """
        A python float representation of the class.
        The returned data is immutable so changes will not mirror the instance.
        """
        return self.value_

    if __major__ <= 2:
        def __add__(self, other):
            warnings.warn(f"__add__ is depreciated on DoubleTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return float(primitive_conversion(self) + primitive_conversion(other))

        def __sub__(self, other):
            warnings.warn(f"__sub__ is depreciated on DoubleTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return float(primitive_conversion(self) - primitive_conversion(other))

        def __mul__(self, other):
            warnings.warn(f"__mul__ is depreciated on DoubleTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return float(primitive_conversion(self) * primitive_conversion(other))

        def __truediv__(self, other):
            warnings.warn(f"__truediv__ is depreciated on DoubleTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return float(primitive_conversion(self) / primitive_conversion(other))

        def __floordiv__(self, other):
            warnings.warn(f"__floordiv__ is depreciated on DoubleTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) // primitive_conversion(other)

        def __mod__(self, other):
            warnings.warn(f"__mod__ is depreciated on DoubleTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) % primitive_conversion(other)

        def __divmod__(self, other):
            warnings.warn(f"__divmod__ is depreciated on DoubleTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return divmod(primitive_conversion(self), primitive_conversion(other))

        def __pow__(self, power, modulo):
            warnings.warn(f"__pow__ is depreciated on DoubleTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return pow(primitive_conversion(self), power, modulo)

        def __lshift__(self, other):
            warnings.warn(f"__lshift__ is depreciated on DoubleTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) << primitive_conversion(other)

        def __rshift__(self, other):
            warnings.warn(f"__rshift__ is depreciated on DoubleTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) >> primitive_conversion(other)

        def __and__(self, other):
            warnings.warn(f"__and__ is depreciated on DoubleTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) & primitive_conversion(other)

        def __xor__(self, other):
            warnings.warn(f"__xor__ is depreciated on DoubleTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) ^ primitive_conversion(other)

        def __or__(self, other):
            warnings.warn(f"__or__ is depreciated on DoubleTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(self) | primitive_conversion(other)

        def __radd__(self, other):
            warnings.warn(f"__radd__ is depreciated on DoubleTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return float(primitive_conversion(other) + primitive_conversion(self))

        def __rsub__(self, other):
            warnings.warn(f"__rsub__ is depreciated on DoubleTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return float(primitive_conversion(other) - primitive_conversion(self))

        def __rmul__(self, other):
            warnings.warn(f"__rmul__ is depreciated on DoubleTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) * primitive_conversion(self)

        def __rtruediv__(self, other):
            warnings.warn(f"__rtruediv__ is depreciated on DoubleTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return float(primitive_conversion(other) / primitive_conversion(self))

        def __rfloordiv__(self, other):
            warnings.warn(f"__rfloordiv__ is depreciated on DoubleTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) // primitive_conversion(self)

        def __rmod__(self, other):
            warnings.warn(f"__rmod__ is depreciated on DoubleTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) % primitive_conversion(self)

        def __rdivmod__(self, other):
            warnings.warn(f"__rdivmod__ is depreciated on DoubleTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return divmod(primitive_conversion(other), primitive_conversion(self))

        def __rpow__(self, other, modulo):
            warnings.warn(f"__rpow__ is depreciated on DoubleTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return pow(primitive_conversion(other), primitive_conversion(self), modulo)

        def __rlshift__(self, other):
            warnings.warn(f"__rlshift__ is depreciated on DoubleTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) << primitive_conversion(self)

        def __rrshift__(self, other):
            warnings.warn(f"__rrshift__ is depreciated on DoubleTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) >> primitive_conversion(self)

        def __rand__(self, other):
            warnings.warn(f"__rand__ is depreciated on DoubleTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) & primitive_conversion(self)

        def __rxor__(self, other):
            warnings.warn(f"__rxor__ is depreciated on DoubleTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) ^ primitive_conversion(self)

        def __ror__(self, other):
            warnings.warn(f"__ror__ is depreciated on DoubleTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return primitive_conversion(other) | primitive_conversion(self)

        def __neg__(self):
            warnings.warn(f"__neg__ is depreciated on DoubleTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return -self.value_

        def __pos__(self):
            warnings.warn(f"__pos__ is depreciated on DoubleTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return +self.value_

        def __abs__(self):
            warnings.warn(f"__abs__ is depreciated on DoubleTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return abs(self.value_)

        def __round__(self, n=None):
            warnings.warn(f"__round__ is depreciated on DoubleTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return round(self.value_, n)

        def __trunc__(self):
            warnings.warn(f"__trunc__ is depreciated on DoubleTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return trunc(self.value_)

        def __floor__(self):
            warnings.warn(f"__floor__ is depreciated on DoubleTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return floor(self.value_)

        def __ceil__(self):
            warnings.warn(f"__ceil__ is depreciated on DoubleTag and will be removed in the future. Please use .py_float to achieve the same behaviour.", DeprecationWarning)
            return ceil(self.value_)

    cdef str _to_snbt(DoubleTag self):
        return f"{self.value_}d"

    cdef void write_payload(
        DoubleTag self,
        object buffer: BytesIO,
        bint little_endian,
        string_encoder: EncoderType,
    ) except *:
        write_double(self.value_, buffer, little_endian)



# Base Types
import warnings
import re

from . import _version

__version__ = _version.get_versions()["version"]
__major__ = int(re.match(r"\d+", __version__)[0])

from ._value import (
    AbstractBaseTag,
    AbstractBaseImmutableTag,
    AbstractBaseMutableTag,
)
from ._numeric import AbstractBaseNumericTag

# Types
from ._int import (
    AbstractBaseIntTag,
    ByteTag,
    ShortTag,
    IntTag,
    LongTag,
    ByteTag as TAG_Byte,
    ShortTag as TAG_Short,
    IntTag as TAG_Int,
    LongTag as TAG_Long,
)
from ._float import (
    AbstractBaseFloatTag,
    FloatTag,
    DoubleTag,
    FloatTag as TAG_Float,
    DoubleTag as TAG_Double,
)
from ._array import (
    AbstractBaseArrayTag,
    ByteArrayTag,
    IntArrayTag,
    LongArrayTag,
    ByteArrayTag as TAG_Byte_Array,
    IntArrayTag as TAG_Int_Array,
    LongArrayTag as TAG_Long_Array,
)
from ._string import (
    StringTag,
    StringTag as TAG_String,
)

from ._named_tag import NamedTag

from ._list import (
    ListTag,
    ListTag as TAG_List,
)
from ._compound import (
    CompoundTag,
    CompoundTag as TAG_Compound,
)

# Load functions
from ._load_nbt import load, load_many, ReadContext
from ._load_snbt import from_snbt

from ._errors import NBTError, NBTLoadError, NBTFormatError, SNBTParseError

from ._dtype import SNBTType, IntType, FloatType, NumberType, ArrayType, AnyNBT

from ._util import utf8_decoder, utf8_encoder, utf8_escape_decoder, utf8_escape_encoder


def __getattr__(name):
    if name == "NBTFile":
        warnings.warn(
            "NBTFile is depreciated. Use NamedTag instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return NamedTag
    elif name == "BaseArrayType":
        warnings.warn(
            "BaseArrayType is depreciated. Use AbstractBaseArrayTag instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return AbstractBaseArrayTag
    elif name == "BaseValueType":
        warnings.warn(
            "BaseValueType is depreciated. Use AbstractBaseTag instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return AbstractBaseTag
    raise AttributeError(f"module {__name__} has no attribute {name}")



## This file is generated by tempita. Do not modify this file directly or your changes will get overwritten.
## To edit this file edit the template in template/src

from typing import List, TypeVar, Generic
from io import BytesIO
from copy import copy, deepcopy
from collections.abc import MutableSequence
import sys
import warnings

from . import __major__
from ._value cimport AbstractBaseTag, AbstractBaseMutableTag
from ._const cimport ID_LIST, CommaSpace, CommaNewline
from ._util cimport write_byte, write_int, BufferContext, read_byte, read_int
if __major__ <= 2:
    from ._util import primitive_conversion
from ._load_nbt cimport load_payload
from ._dtype import AnyNBT, EncoderType


cdef inline CyListTag read_list_tag(BufferContext buffer, bint little_endian, string_decoder: DecoderType):
    cdef CyListTag tag = ListTag()
    cdef char list_type = read_byte(buffer)
    tag.list_data_type = list_type
    cdef int length = read_int(buffer, little_endian)
    cdef list val = tag.value_
    cdef int i
    for i in range(length):
        val.append(load_payload(buffer, list_type, little_endian, string_decoder))
    return tag


cdef class CyListTag(AbstractBaseMutableTag):
    """
    This class behaves like a python list.
    All contained data must be of the same NBT data type.
    """
    tag_id = ID_LIST

    def __init__(CyListTag self, object value = (), char list_data_type = 1):
        self.list_data_type = list_data_type
        self.value_ = []
        cdef list list_value
        list_value = list(value)
        self._check_tag_iterable(list_value)
        self.value_ = list_value

    @classmethod
    def create(cls, object value = (), char list_data_type = 1) -> ListTag:
        return ListTag(value, list_data_type)

    def __str__(CyListTag self):
        return str(self.value_)

    def __eq__(CyListTag self, other):
        cdef CyListTag other_
        if isinstance(other, CyListTag):
            other_ = other
            return self.value_ == other_.value_
        elif __major__ <= 2:
            warnings.warn("NBT comparison operator (a == b) will only return True between classes of the same type.", FutureWarning)
            return self.value_ == primitive_conversion(other)
        return NotImplemented

    def __reduce__(CyListTag self):
        return self.__class__, (self.value_,)

    def __deepcopy__(CyListTag self, memo=None):
        return self.__class__(deepcopy(self.value_, memo=memo))

    def __copy__(CyListTag self):
        return self.__class__(self.value_)

    @property
    def py_list(CyListTag self) -> List[AnyNBT]:
        """
        A python list representation of the class.
        The returned list is a shallow copy of the class, meaning changes will not mirror the instance.
        Use the public API to modify the internal data.
        """
        return copy(self.value_)

    @property
    def py_data(self):
        """
        A python representation of the class. Note that the return type is undefined and may change in the future.
        You would be better off using the py_{type} or np_array properties if you require a fixed type.
        This is here for convenience to get a python representation under the same property name.
        """
        return self.py_list

    cdef void _check_tag(CyListTag self, AbstractBaseTag value, bint fix_if_empty=True) except *:
        if value is None:
            raise TypeError("List values must be NBT Tags")
        if fix_if_empty and not self.value_:
            self.list_data_type = value.tag_id
        elif value.tag_id != self.list_data_type:
            raise TypeError(
                f"Invalid type {value.__class__.__name__} for ListTag(list_data_type={self.list_data_type})"
            )

    cdef void _check_tag_iterable(CyListTag self, list value) except *:
        cdef int i
        cdef AbstractBaseTag tag
        for i, tag in enumerate(value):
            self._check_tag(tag, not i)

    cdef void write_payload(
        CyListTag self,
        object buffer: BytesIO,
        bint little_endian,
        string_encoder: EncoderType,
    ) except *:
        cdef char list_type = self.list_data_type

        write_byte(list_type, buffer)
        write_int(<int> len(self.value_), buffer, little_endian)

        cdef AbstractBaseTag subtag
        for subtag in self.value_:
            if subtag.tag_id != list_type:
                raise ValueError(
                    f"ListTag must only contain one type! Found {subtag.tag_id} in list type {list_type}"
                )
            subtag.write_payload(buffer, little_endian, string_encoder)

    cdef str _to_snbt(CyListTag self):
        cdef AbstractBaseTag elem
        cdef list tags = []
        for elem in self.value_:
            tags.append(elem._to_snbt())
        return f"[{CommaSpace.join(tags)}]"

    cdef str _pretty_to_snbt(CyListTag self, str indent_chr, int indent_count=0, bint leading_indent=True):
        cdef AbstractBaseTag elem
        cdef list tags = []
        for elem in self.value_:
            tags.append(elem._pretty_to_snbt(indent_chr, indent_count + 1))
        if tags:
            return f"{indent_chr * indent_count * leading_indent}[\n{CommaNewline.join(tags)}\n{indent_chr * indent_count}]"
        else:
            return f"{indent_chr * indent_count * leading_indent}[]"

    def __repr__(CyListTag self):
        return f"{self.__class__.__name__}({repr(self.value_)}, {self.list_data_type})"

    def __contains__(CyListTag self, object item) -> bool:
        return isinstance(item, AbstractBaseTag) and item.tag_id == self.list_data_type and self.value_.__contains__(item)

    def __len__(CyListTag self) -> int:
        return self.value_.__len__()

    def __getitem__(CyListTag self, index) -> AnyNBT:
        return self.value_[index]

    def __setitem__(CyListTag self, index, value):
        if isinstance(index, slice):
            value = list(value)
            self._check_tag_iterable(value)
        else:
            self._check_tag(value)
        self.value_[index] = value

    def __delitem__(CyListTag self, object index):
        self.value_.__delitem__(index)

    def copy(CyListTag self):
        """Return a shallow copy of the class"""
        return ListTag(self.value_.copy(), self.list_data_type)

    def insert(CyListTag self, object index, AbstractBaseTag value not None):
        self._check_tag(value)
        self.value_.insert(index, value)

    cpdef ByteTag get_byte(self, int index):
        """Get the tag at index if it is a ByteTag.
    
        :param index: The index to get
        :return: The ByteTag.
        :raises: IndexError if the index is outside the list.
        :raises: TypeError if the stored type is not a ByteTag
        """
        cdef ByteTag tag = self[index]
        return tag

    cpdef ShortTag get_short(self, int index):
        """Get the tag at index if it is a ShortTag.
    
        :param index: The index to get
        :return: The ShortTag.
        :raises: IndexError if the index is outside the list.
        :raises: TypeError if the stored type is not a ShortTag
        """
        cdef ShortTag tag = self[index]
        return tag

    cpdef IntTag get_int(self, int index):
        """Get the tag at index if it is a IntTag.
    
        :param index: The index to get
        :return: The IntTag.
        :raises: IndexError if the index is outside the list.
        :raises: TypeError if the stored type is not a IntTag
        """
        cdef IntTag tag = self[index]
        return tag

    cpdef LongTag get_long(self, int index):
        """Get the tag at index if it is a LongTag.
    
        :param index: The index to get
        :return: The LongTag.
        :raises: IndexError if the index is outside the list.
        :raises: TypeError if the stored type is not a LongTag
        """
        cdef LongTag tag = self[index]
        return tag

    cpdef FloatTag get_float(self, int index):
        """Get the tag at index if it is a FloatTag.
    
        :param index: The index to get
        :return: The FloatTag.
        :raises: IndexError if the index is outside the list.
        :raises: TypeError if the stored type is not a FloatTag
        """
        cdef FloatTag tag = self[index]
        return tag

    cpdef DoubleTag get_double(self, int index):
        """Get the tag at index if it is a DoubleTag.
    
        :param index: The index to get
        :return: The DoubleTag.
        :raises: IndexError if the index is outside the list.
        :raises: TypeError if the stored type is not a DoubleTag
        """
        cdef DoubleTag tag = self[index]
        return tag

    cpdef StringTag get_string(self, int index):
        """Get the tag at index if it is a StringTag.
    
        :param index: The index to get
        :return: The StringTag.
        :raises: IndexError if the index is outside the list.
        :raises: TypeError if the stored type is not a StringTag
        """
        cdef StringTag tag = self[index]
        return tag

    cpdef CyListTag get_list(self, int index):
        """Get the tag at index if it is a CyListTag.
    
        :param index: The index to get
        :return: The CyListTag.
        :raises: IndexError if the index is outside the list.
        :raises: TypeError if the stored type is not a CyListTag
        """
        cdef CyListTag tag = self[index]
        return tag

    cpdef CyCompoundTag get_compound(self, int index):
        """Get the tag at index if it is a CyCompoundTag.
    
        :param index: The index to get
        :return: The CyCompoundTag.
        :raises: IndexError if the index is outside the list.
        :raises: TypeError if the stored type is not a CyCompoundTag
        """
        cdef CyCompoundTag tag = self[index]
        return tag

    cpdef ByteArrayTag get_byte_array(self, int index):
        """Get the tag at index if it is a ByteArrayTag.
    
        :param index: The index to get
        :return: The ByteArrayTag.
        :raises: IndexError if the index is outside the list.
        :raises: TypeError if the stored type is not a ByteArrayTag
        """
        cdef ByteArrayTag tag = self[index]
        return tag

    cpdef IntArrayTag get_int_array(self, int index):
        """Get the tag at index if it is a IntArrayTag.
    
        :param index: The index to get
        :return: The IntArrayTag.
        :raises: IndexError if the index is outside the list.
        :raises: TypeError if the stored type is not a IntArrayTag
        """
        cdef IntArrayTag tag = self[index]
        return tag

    cpdef LongArrayTag get_long_array(self, int index):
        """Get the tag at index if it is a LongArrayTag.
    
        :param index: The index to get
        :return: The LongArrayTag.
        :raises: IndexError if the index is outside the list.
        :raises: TypeError if the stored type is not a LongArrayTag
        """
        cdef LongArrayTag tag = self[index]
        return tag


class ListTag(CyListTag, MutableSequence):
    pass



from ._numeric cimport AbstractBaseNumericTag
from ._util cimport BufferContext

cdef class AbstractBaseFloatTag(AbstractBaseNumericTag):
    pass

cdef FloatTag read_float_tag(BufferContext buffer, bint little_endian)

cdef class FloatTag(AbstractBaseFloatTag):
    cdef float value_

cdef DoubleTag read_double_tag(BufferContext buffer, bint little_endian)

cdef class DoubleTag(AbstractBaseFloatTag):
    cdef double value_



import numpy
cimport numpy
numpy.import_array()
from ._value cimport AbstractBaseMutableTag
from ._util cimport BufferContext

cdef class AbstractBaseArrayTag(AbstractBaseMutableTag):
    pass

cdef ByteArrayTag read_byte_array_tag(BufferContext buffer, bint little_endian)

cdef class ByteArrayTag(AbstractBaseArrayTag):
    cdef numpy.ndarray value_

cdef IntArrayTag read_int_array_tag(BufferContext buffer, bint little_endian)

cdef class IntArrayTag(AbstractBaseArrayTag):
    cdef numpy.ndarray value_

cdef LongArrayTag read_long_array_tag(BufferContext buffer, bint little_endian)

cdef class LongArrayTag(AbstractBaseArrayTag):
    cdef numpy.ndarray value_



## This file is generated by tempita. Do not modify this file directly or your changes will get overwritten.
## To edit this file edit the template in template/src

from io import BytesIO
import re
from typing import Iterator, Dict, Type, TypeVar
from copy import copy, deepcopy
from collections.abc import MutableMapping
import sys
import warnings

from . import __major__
from ._value cimport AbstractBaseTag, AbstractBaseMutableTag
from ._const cimport ID_END, ID_COMPOUND, CommaSpace, CommaNewline
from ._util cimport write_byte, BufferContext, read_byte, read_string
if __major__ <= 2:
    from ._util import primitive_conversion
from ._load_nbt cimport load_payload
from ._dtype import AnyNBT, DecoderType, EncoderType
from ._int cimport (
    ByteTag,
    ShortTag,
    IntTag,
    LongTag,
)
from ._float cimport (
    FloatTag,
    DoubleTag,
)
from ._array cimport (
    ByteArrayTag,
    IntArrayTag,
    LongArrayTag,
)
from ._string cimport StringTag
from ._list cimport CyListTag

NON_QUOTED_KEY = re.compile('[A-Za-z0-9._+-]+')

T = TypeVar("T")


cdef inline CyCompoundTag read_compound_tag(BufferContext buffer, bint little_endian, string_decoder: DecoderType):
    cdef CyCompoundTag tag = CompoundTag()
    cdef char tag_type
    cdef str name
    cdef AbstractBaseTag child_tag

    while True:
        tag_type = read_byte(buffer)
        if tag_type == ID_END:
            break
        else:
            name = read_string(buffer, little_endian, string_decoder)
            child_tag = load_payload(buffer, tag_type, little_endian, string_decoder)
            tag[name] = child_tag
    return tag


cdef inline void _check_dict(dict value) except *:
    cdef str key
    cdef AbstractBaseTag val
    for key, val in value.items():
        if key is None or val is None:
            raise TypeError()


cdef class CyCompoundTag(AbstractBaseMutableTag):
    """
    This class behaves like a python dictionary.
    All keys must be strings and all values must be NBT data types.
    """
    tag_id = ID_COMPOUND

    def __init__(CyCompoundTag self, object value = (), **kwargs):
        cdef dict dict_value = dict(value, **kwargs)
        _check_dict(dict_value)
        self.value_ = dict_value

    @classmethod
    def create(cls, object value = (), **kwargs) -> CompoundTag:
        return CompoundTag(value, **kwargs)

    def __str__(CyCompoundTag self):
        return str(self.value_)

    def __eq__(CyCompoundTag self, other):
        cdef CyCompoundTag other_
        if isinstance(other, CyCompoundTag):
            other_ = other
            return self.value_ == other_.value_
        elif __major__ <= 2:
            warnings.warn("NBT comparison operator (a == b) will only return True between classes of the same type.", FutureWarning)
            return self.value_ == primitive_conversion(other)
        return NotImplemented

    def __reduce__(CyCompoundTag self):
        return self.__class__, (self.value_,)

    def __deepcopy__(CyCompoundTag self, memo=None):
        return self.__class__(deepcopy(self.value_, memo=memo))

    def __copy__(CyCompoundTag self):
        return self.__class__(self.value_)

    @property
    def py_dict(CyCompoundTag self) -> Dict[str, AnyNBT]:
        """
        A python dictionary representation of the class.
        The returned list is a shallow copy of the class, meaning changes will not mirror the instance.
        Use the public API to modify the internal data.
        """
        return copy(self.value_)

    @property
    def py_data(self):
        """
        A python representation of the class. Note that the return type is undefined and may change in the future.
        You would be better off using the py_{type} or np_array properties if you require a fixed type.
        This is here for convenience to get a python representation under the same property name.
        """
        return self.py_dict

    @staticmethod
    def fromkeys(object keys, AbstractBaseTag value=None):
        return CompoundTag(dict.fromkeys(keys, value))

    cdef str _to_snbt(CyCompoundTag self):
        cdef str name
        cdef AbstractBaseTag elem
        cdef list tags = []
        for name in sorted(self.value_, key=lambda k: (k.lower(), k.swapcase())):
            elem = self.value_[name]
            if NON_QUOTED_KEY.fullmatch(name) is None:
                tags.append(f'"{name}": {elem.to_snbt()}')
            else:
                tags.append(f'{name}: {elem.to_snbt()}')
        return f"{{{CommaSpace.join(tags)}}}"

    cdef str _pretty_to_snbt(CyCompoundTag self, str indent_chr, int indent_count=0, bint leading_indent=True):
        cdef str name
        cdef AbstractBaseTag elem
        cdef list tags = []
        for name in sorted(self.value_, key=lambda k: (k.lower(), k.swapcase())):
            elem = self.value_[name]
            tags.append(f'{indent_chr * (indent_count + 1)}"{name}": {elem._pretty_to_snbt(indent_chr, indent_count + 1, False)}')
        if tags:
            return f"{indent_chr * indent_count * leading_indent}{{\n{CommaNewline.join(tags)}\n{indent_chr * indent_count}}}"
        else:
            return f"{indent_chr * indent_count * leading_indent}{{}}"

    cdef void write_payload(
        CyCompoundTag self,
        object buffer: BytesIO,
        bint little_endian,
        string_encoder: EncoderType
    ) except *:
        cdef str key
        cdef AbstractBaseTag tag

        for key, tag in self.value_.items():
            tag.write_tag(buffer, key, little_endian, string_encoder)
        write_byte(ID_END, buffer)

    def __repr__(CyCompoundTag self):
        return f"{self.__class__.__name__}({repr(self.value_)})"

    def __getitem__(CyCompoundTag self, str key not None) -> AbstractBaseTag:
        return self.value_[key]

    def __setitem__(CyCompoundTag self, str key not None, AbstractBaseTag value not None):
        self.value_[key] = value

    def __delitem__(CyCompoundTag self, str key not None):
        self.value_.__delitem__(key)

    def __iter__(CyCompoundTag self) -> Iterator[str]:
        yield from self.value_

    def __len__(CyCompoundTag self) -> int:
        return self.value_.__len__()

    cpdef ByteTag get_byte(self, str key, ByteTag default=None):
        """Get the tag stored in key if it is a ByteTag.
    
        :param key: The key to get
        :param default: The value to return if the key does not exist or the type is wrong. If not defined errors are raised.
        :return: The ByteTag.
        :raises: KeyError if the key does not exist
        :raises: TypeError if the stored type is not a ByteTag
        """
        if key not in self:
            if default is None:
                raise KeyError(key)
            else:
                return default

        tag = self[key]
        if isinstance(tag, ByteTag):
            return tag
        elif default is None:
            raise TypeError(f"Expected tag to be of type ByteTag but got {type(tag)}")
        else:
            return default

    cpdef ByteTag setdefault_byte(self, str key, ByteTag default=None):
        """Populate key if not defined or value is not ByteTag. Return the value stored.
    
        If default is a ByteTag then it will be stored under key else a default instance will be created.
        :param key: The key to populate and get
        :param default: The default value to use
        :return: The ByteTag stored in key
        :raises: TypeError if the input types are incorrect
        """
        val = self[key] if key in self else None
        if not isinstance(val, ByteTag):
            val = self[key] = ByteTag() if default is None else default
        return val

    cpdef ShortTag get_short(self, str key, ShortTag default=None):
        """Get the tag stored in key if it is a ShortTag.
    
        :param key: The key to get
        :param default: The value to return if the key does not exist or the type is wrong. If not defined errors are raised.
        :return: The ShortTag.
        :raises: KeyError if the key does not exist
        :raises: TypeError if the stored type is not a ShortTag
        """
        if key not in self:
            if default is None:
                raise KeyError(key)
            else:
                return default

        tag = self[key]
        if isinstance(tag, ShortTag):
            return tag
        elif default is None:
            raise TypeError(f"Expected tag to be of type ShortTag but got {type(tag)}")
        else:
            return default

    cpdef ShortTag setdefault_short(self, str key, ShortTag default=None):
        """Populate key if not defined or value is not ShortTag. Return the value stored.
    
        If default is a ShortTag then it will be stored under key else a default instance will be created.
        :param key: The key to populate and get
        :param default: The default value to use
        :return: The ShortTag stored in key
        :raises: TypeError if the input types are incorrect
        """
        val = self[key] if key in self else None
        if not isinstance(val, ShortTag):
            val = self[key] = ShortTag() if default is None else default
        return val

    cpdef IntTag get_int(self, str key, IntTag default=None):
        """Get the tag stored in key if it is a IntTag.
    
        :param key: The key to get
        :param default: The value to return if the key does not exist or the type is wrong. If not defined errors are raised.
        :return: The IntTag.
        :raises: KeyError if the key does not exist
        :raises: TypeError if the stored type is not a IntTag
        """
        if key not in self:
            if default is None:
                raise KeyError(key)
            else:
                return default

        tag = self[key]
        if isinstance(tag, IntTag):
            return tag
        elif default is None:
            raise TypeError(f"Expected tag to be of type IntTag but got {type(tag)}")
        else:
            return default

    cpdef IntTag setdefault_int(self, str key, IntTag default=None):
        """Populate key if not defined or value is not IntTag. Return the value stored.
    
        If default is a IntTag then it will be stored under key else a default instance will be created.
        :param key: The key to populate and get
        :param default: The default value to use
        :return: The IntTag stored in key
        :raises: TypeError if the input types are incorrect
        """
        val = self[key] if key in self else None
        if not isinstance(val, IntTag):
            val = self[key] = IntTag() if default is None else default
        return val

    cpdef LongTag get_long(self, str key, LongTag default=None):
        """Get the tag stored in key if it is a LongTag.
    
        :param key: The key to get
        :param default: The value to return if the key does not exist or the type is wrong. If not defined errors are raised.
        :return: The LongTag.
        :raises: KeyError if the key does not exist
        :raises: TypeError if the stored type is not a LongTag
        """
        if key not in self:
            if default is None:
                raise KeyError(key)
            else:
                return default

        tag = self[key]
        if isinstance(tag, LongTag):
            return tag
        elif default is None:
            raise TypeError(f"Expected tag to be of type LongTag but got {type(tag)}")
        else:
            return default

    cpdef LongTag setdefault_long(self, str key, LongTag default=None):
        """Populate key if not defined or value is not LongTag. Return the value stored.
    
        If default is a LongTag then it will be stored under key else a default instance will be created.
        :param key: The key to populate and get
        :param default: The default value to use
        :return: The LongTag stored in key
        :raises: TypeError if the input types are incorrect
        """
        val = self[key] if key in self else None
        if not isinstance(val, LongTag):
            val = self[key] = LongTag() if default is None else default
        return val

    cpdef FloatTag get_float(self, str key, FloatTag default=None):
        """Get the tag stored in key if it is a FloatTag.
    
        :param key: The key to get
        :param default: The value to return if the key does not exist or the type is wrong. If not defined errors are raised.
        :return: The FloatTag.
        :raises: KeyError if the key does not exist
        :raises: TypeError if the stored type is not a FloatTag
        """
        if key not in self:
            if default is None:
                raise KeyError(key)
            else:
                return default

        tag = self[key]
        if isinstance(tag, FloatTag):
            return tag
        elif default is None:
            raise TypeError(f"Expected tag to be of type FloatTag but got {type(tag)}")
        else:
            return default

    cpdef FloatTag setdefault_float(self, str key, FloatTag default=None):
        """Populate key if not defined or value is not FloatTag. Return the value stored.
    
        If default is a FloatTag then it will be stored under key else a default instance will be created.
        :param key: The key to populate and get
        :param default: The default value to use
        :return: The FloatTag stored in key
        :raises: TypeError if the input types are incorrect
        """
        val = self[key] if key in self else None
        if not isinstance(val, FloatTag):
            val = self[key] = FloatTag() if default is None else default
        return val

    cpdef DoubleTag get_double(self, str key, DoubleTag default=None):
        """Get the tag stored in key if it is a DoubleTag.
    
        :param key: The key to get
        :param default: The value to return if the key does not exist or the type is wrong. If not defined errors are raised.
        :return: The DoubleTag.
        :raises: KeyError if the key does not exist
        :raises: TypeError if the stored type is not a DoubleTag
        """
        if key not in self:
            if default is None:
                raise KeyError(key)
            else:
                return default

        tag = self[key]
        if isinstance(tag, DoubleTag):
            return tag
        elif default is None:
            raise TypeError(f"Expected tag to be of type DoubleTag but got {type(tag)}")
        else:
            return default

    cpdef DoubleTag setdefault_double(self, str key, DoubleTag default=None):
        """Populate key if not defined or value is not DoubleTag. Return the value stored.
    
        If default is a DoubleTag then it will be stored under key else a default instance will be created.
        :param key: The key to populate and get
        :param default: The default value to use
        :return: The DoubleTag stored in key
        :raises: TypeError if the input types are incorrect
        """
        val = self[key] if key in self else None
        if not isinstance(val, DoubleTag):
            val = self[key] = DoubleTag() if default is None else default
        return val

    cpdef StringTag get_string(self, str key, StringTag default=None):
        """Get the tag stored in key if it is a StringTag.
    
        :param key: The key to get
        :param default: The value to return if the key does not exist or the type is wrong. If not defined errors are raised.
        :return: The StringTag.
        :raises: KeyError if the key does not exist
        :raises: TypeError if the stored type is not a StringTag
        """
        if key not in self:
            if default is None:
                raise KeyError(key)
            else:
                return default

        tag = self[key]
        if isinstance(tag, StringTag):
            return tag
        elif default is None:
            raise TypeError(f"Expected tag to be of type StringTag but got {type(tag)}")
        else:
            return default

    cpdef StringTag setdefault_string(self, str key, StringTag default=None):
        """Populate key if not defined or value is not StringTag. Return the value stored.
    
        If default is a StringTag then it will be stored under key else a default instance will be created.
        :param key: The key to populate and get
        :param default: The default value to use
        :return: The StringTag stored in key
        :raises: TypeError if the input types are incorrect
        """
        val = self[key] if key in self else None
        if not isinstance(val, StringTag):
            val = self[key] = StringTag() if default is None else default
        return val

    cpdef CyListTag get_list(self, str key, CyListTag default=None):
        """Get the tag stored in key if it is a CyListTag.
    
        :param key: The key to get
        :param default: The value to return if the key does not exist or the type is wrong. If not defined errors are raised.
        :return: The CyListTag.
        :raises: KeyError if the key does not exist
        :raises: TypeError if the stored type is not a CyListTag
        """
        if key not in self:
            if default is None:
                raise KeyError(key)
            else:
                return default

        tag = self[key]
        if isinstance(tag, CyListTag):
            return tag
        elif default is None:
            raise TypeError(f"Expected tag to be of type CyListTag but got {type(tag)}")
        else:
            return default

    cpdef CyListTag setdefault_list(self, str key, CyListTag default=None):
        """Populate key if not defined or value is not CyListTag. Return the value stored.
    
        If default is a CyListTag then it will be stored under key else a default instance will be created.
        :param key: The key to populate and get
        :param default: The default value to use
        :return: The CyListTag stored in key
        :raises: TypeError if the input types are incorrect
        """
        val = self[key] if key in self else None
        if not isinstance(val, CyListTag):
            val = self[key] = CyListTag.create() if default is None else default
        return val

    cpdef CyCompoundTag get_compound(self, str key, CyCompoundTag default=None):
        """Get the tag stored in key if it is a CyCompoundTag.
    
        :param key: The key to get
        :param default: The value to return if the key does not exist or the type is wrong. If not defined errors are raised.
        :return: The CyCompoundTag.
        :raises: KeyError if the key does not exist
        :raises: TypeError if the stored type is not a CyCompoundTag
        """
        if key not in self:
            if default is None:
                raise KeyError(key)
            else:
                return default

        tag = self[key]
        if isinstance(tag, CyCompoundTag):
            return tag
        elif default is None:
            raise TypeError(f"Expected tag to be of type CyCompoundTag but got {type(tag)}")
        else:
            return default

    cpdef CyCompoundTag setdefault_compound(self, str key, CyCompoundTag default=None):
        """Populate key if not defined or value is not CyCompoundTag. Return the value stored.
    
        If default is a CyCompoundTag then it will be stored under key else a default instance will be created.
        :param key: The key to populate and get
        :param default: The default value to use
        :return: The CyCompoundTag stored in key
        :raises: TypeError if the input types are incorrect
        """
        val = self[key] if key in self else None
        if not isinstance(val, CyCompoundTag):
            val = self[key] = CompoundTag() if default is None else default
        return val

    cpdef ByteArrayTag get_byte_array(self, str key, ByteArrayTag default=None):
        """Get the tag stored in key if it is a ByteArrayTag.
    
        :param key: The key to get
        :param default: The value to return if the key does not exist or the type is wrong. If not defined errors are raised.
        :return: The ByteArrayTag.
        :raises: KeyError if the key does not exist
        :raises: TypeError if the stored type is not a ByteArrayTag
        """
        if key not in self:
            if default is None:
                raise KeyError(key)
            else:
                return default

        tag = self[key]
        if isinstance(tag, ByteArrayTag):
            return tag
        elif default is None:
            raise TypeError(f"Expected tag to be of type ByteArrayTag but got {type(tag)}")
        else:
            return default

    cpdef ByteArrayTag setdefault_byte_array(self, str key, ByteArrayTag default=None):
        """Populate key if not defined or value is not ByteArrayTag. Return the value stored.
    
        If default is a ByteArrayTag then it will be stored under key else a default instance will be created.
        :param key: The key to populate and get
        :param default: The default value to use
        :return: The ByteArrayTag stored in key
        :raises: TypeError if the input types are incorrect
        """
        val = self[key] if key in self else None
        if not isinstance(val, ByteArrayTag):
            val = self[key] = ByteArrayTag() if default is None else default
        return val

    cpdef IntArrayTag get_int_array(self, str key, IntArrayTag default=None):
        """Get the tag stored in key if it is a IntArrayTag.
    
        :param key: The key to get
        :param default: The value to return if the key does not exist or the type is wrong. If not defined errors are raised.
        :return: The IntArrayTag.
        :raises: KeyError if the key does not exist
        :raises: TypeError if the stored type is not a IntArrayTag
        """
        if key not in self:
            if default is None:
                raise KeyError(key)
            else:
                return default

        tag = self[key]
        if isinstance(tag, IntArrayTag):
            return tag
        elif default is None:
            raise TypeError(f"Expected tag to be of type IntArrayTag but got {type(tag)}")
        else:
            return default

    cpdef IntArrayTag setdefault_int_array(self, str key, IntArrayTag default=None):
        """Populate key if not defined or value is not IntArrayTag. Return the value stored.
    
        If default is a IntArrayTag then it will be stored under key else a default instance will be created.
        :param key: The key to populate and get
        :param default: The default value to use
        :return: The IntArrayTag stored in key
        :raises: TypeError if the input types are incorrect
        """
        val = self[key] if key in self else None
        if not isinstance(val, IntArrayTag):
            val = self[key] = IntArrayTag() if default is None else default
        return val

    cpdef LongArrayTag get_long_array(self, str key, LongArrayTag default=None):
        """Get the tag stored in key if it is a LongArrayTag.
    
        :param key: The key to get
        :param default: The value to return if the key does not exist or the type is wrong. If not defined errors are raised.
        :return: The LongArrayTag.
        :raises: KeyError if the key does not exist
        :raises: TypeError if the stored type is not a LongArrayTag
        """
        if key not in self:
            if default is None:
                raise KeyError(key)
            else:
                return default

        tag = self[key]
        if isinstance(tag, LongArrayTag):
            return tag
        elif default is None:
            raise TypeError(f"Expected tag to be of type LongArrayTag but got {type(tag)}")
        else:
            return default

    cpdef LongArrayTag setdefault_long_array(self, str key, LongArrayTag default=None):
        """Populate key if not defined or value is not LongArrayTag. Return the value stored.
    
        If default is a LongArrayTag then it will be stored under key else a default instance will be created.
        :param key: The key to populate and get
        :param default: The default value to use
        :return: The LongArrayTag stored in key
        :raises: TypeError if the input types are incorrect
        """
        val = self[key] if key in self else None
        if not isinstance(val, LongArrayTag):
            val = self[key] = LongArrayTag() if default is None else default
        return val


class CompoundTag(CyCompoundTag, MutableMapping):
    pass



from ._value cimport AbstractBaseTag, AbstractBaseMutableTag
from ._util cimport BufferContext
from ._dtype import DecoderType
from ._int cimport (
    ByteTag,
    ShortTag,
    IntTag,
    LongTag,
)
from ._float cimport (
    FloatTag,
    DoubleTag,
)
from ._array cimport (
    ByteArrayTag,
    IntArrayTag,
    LongArrayTag,
)
from ._string cimport StringTag
from ._compound cimport CyCompoundTag


cdef CyListTag read_list_tag(BufferContext buffer, bint little_endian, string_decoder: DecoderType)

cdef class CyListTag(AbstractBaseMutableTag):
    cdef list value_
    cdef readonly char list_data_type

    cdef void _check_tag(self, AbstractBaseTag value, bint fix_if_empty=*) except *
    cdef void _check_tag_iterable(self, list value) except *

    cpdef ByteTag get_byte(self, int index)
    cpdef ShortTag get_short(self, int index)
    cpdef IntTag get_int(self, int index)
    cpdef LongTag get_long(self, int index)
    cpdef FloatTag get_float(self, int index)
    cpdef DoubleTag get_double(self, int index)
    cpdef StringTag get_string(self, int index)
    cpdef CyListTag get_list(self, int index)
    cpdef CyCompoundTag get_compound(self, int index)
    cpdef ByteArrayTag get_byte_array(self, int index)
    cpdef IntArrayTag get_int_array(self, int index)
    cpdef LongArrayTag get_long_array(self, int index)



from typing import Type, TypeVar
import warnings

from mutf8 import encode_modified_utf8

from . import __major__
from ._value cimport AbstractBaseTag, AbstractBase
from ._int cimport (
    ByteTag,
    ShortTag,
    IntTag,
    LongTag,
)
from ._float cimport (
    FloatTag,
    DoubleTag,
)
from ._array cimport (
    ByteArrayTag,
    IntArrayTag,
    LongArrayTag,
)
from ._string cimport StringTag
from ._list cimport CyListTag
from ._compound cimport CyCompoundTag
from ._dtype import EncoderType, AnyNBT


T = TypeVar("T")


cdef class NamedTag(AbstractBase):
    def __init__(self, AbstractBaseTag tag = None, str name not None = ""):
        if tag is None:
            self.tag = CyCompoundTag.create()
        else:
            self.tag = tag
        self.name = name

    if __major__ <= 2:
        @property
        def value(self):
            """Legacy property to get the tag. Depreciated. Use :attr:`tag` or get_{type} instead."""
            warnings.warn("value property is depreciated. Use tag attribute instead.", DeprecationWarning)
            return self.tag

        @value.setter
        def value(self, AbstractBaseTag value):
            warnings.warn("value property is depreciated.", DeprecationWarning)
            self.tag = value

    @property
    def byte(self) -> ByteTag:
        """Get the tag if it is a ByteTag.

        :return: The ByteTag.
        :raises: TypeError if the stored type is not a ByteTag
        """
        cdef ByteTag tag = self.tag
        return tag

    @property
    def short(self) -> ShortTag:
        """Get the tag if it is a ShortTag.

        :return: The ShortTag.
        :raises: TypeError if the stored type is not a ShortTag
        """
        cdef ShortTag tag = self.tag
        return tag

    @property
    def int(self) -> IntTag:
        """Get the tag if it is a IntTag.

        :return: The IntTag.
        :raises: TypeError if the stored type is not a IntTag
        """
        cdef IntTag tag = self.tag
        return tag

    @property
    def long(self) -> LongTag:
        """Get the tag if it is a LongTag.

        :return: The LongTag.
        :raises: TypeError if the stored type is not a LongTag
        """
        cdef LongTag tag = self.tag
        return tag

    @property
    def float(self) -> FloatTag:
        """Get the tag if it is a FloatTag.

        :return: The FloatTag.
        :raises: TypeError if the stored type is not a FloatTag
        """
        cdef FloatTag tag = self.tag
        return tag

    @property
    def double(self) -> DoubleTag:
        """Get the tag if it is a DoubleTag.

        :return: The DoubleTag.
        :raises: TypeError if the stored type is not a DoubleTag
        """
        cdef DoubleTag tag = self.tag
        return tag

    @property
    def string(self) -> StringTag:
        """Get the tag if it is a StringTag.

        :return: The StringTag.
        :raises: TypeError if the stored type is not a StringTag
        """
        cdef StringTag tag = self.tag
        return tag

    @property
    def list(self) -> CyListTag:
        """Get the tag if it is a CyListTag.

        :return: The CyListTag.
        :raises: TypeError if the stored type is not a CyListTag
        """
        cdef CyListTag tag = self.tag
        return tag

    @property
    def compound(self) -> CyCompoundTag:
        """Get the tag if it is a CyCompoundTag.

        :return: The CyCompoundTag.
        :raises: TypeError if the stored type is not a CyCompoundTag
        """
        cdef CyCompoundTag tag = self.tag
        return tag

    @property
    def byte_array(self) -> ByteArrayTag:
        """Get the tag if it is a ByteArrayTag.

        :return: The ByteArrayTag.
        :raises: TypeError if the stored type is not a ByteArrayTag
        """
        cdef ByteArrayTag tag = self.tag
        return tag

    @property
    def int_array(self) -> IntArrayTag:
        """Get the tag if it is a IntArrayTag.

        :return: The IntArrayTag.
        :raises: TypeError if the stored type is not a IntArrayTag
        """
        cdef IntArrayTag tag = self.tag
        return tag

    @property
    def long_array(self) -> LongArrayTag:
        """Get the tag if it is a LongArrayTag.

        :return: The LongArrayTag.
        :raises: TypeError if the stored type is not a LongArrayTag
        """
        cdef LongArrayTag tag = self.tag
        return tag

    cpdef str to_snbt(self, object indent=None, object indent_chr=None):
        return self.tag.to_snbt(indent=indent, indent_chr=indent_chr)

    def to_nbt(
        self,
        *,
        bint compressed=True,
        bint little_endian=False,
        string_encoder: EncoderType = encode_modified_utf8,
    ) -> bytes:
        return self.tag.to_nbt(
            compressed=compressed,
            little_endian=little_endian,
            string_encoder=string_encoder,
            name = self.name,
        )

    def save_to(
        self,
        object filepath_or_buffer=None,
        *args,
        bint compressed=True,
        bint little_endian=False,
        string_encoder: EncoderType = encode_modified_utf8,
    ) -> bytes:
        if args:
            if __major__ <= 2:
                warnings.warn("save_to arguments are going to become keyword only", FutureWarning)
                compressed, *args = args
                if args:
                    little_endian, *args = args
            else:
                raise TypeError(f"save_to() takes 1 to 2 positional arguments but {2+len(args)} were given")
        return self.tag.save_to(
            filepath_or_buffer,
            compressed=compressed,
            little_endian=little_endian,
            string_encoder=string_encoder,
            name=self.name,
        )

    def __eq__(self, other):
        if isinstance(other, NamedTag):
            return self.name == other.name and self.tag == other.tag
        return NotImplemented

    def __repr__(self):
        return f'NamedTag({repr(self.tag)}, "{self.name}")'

    def __getitem__(self, object item):
        if isinstance(item, int):
            return (self.name, self.tag)[item]
        elif isinstance(item, str):
            warnings.warn("The behaviour of __getitem__ is changing to make NamedTag behave like a named tuple. access the tag to get the same behaviour", DeprecationWarning)
            return self.compound[item]
        else:
            raise TypeError("Item must be an int")

    # def __getitem__(self, int item):
    #     return (self.name, self.tag)[item]

    if __major__ >= 3:
        def __iter__(self):
            yield self.name
            yield self.tag

    else:
        def __iter__(self):
            warnings.warn("iter behaviour on NamedTag is changing. Use the compound property to implement the same behaviour.", FutureWarning)
            yield from self.compound

        def __len__(self) -> int:
            """Depreciated. Do not use."""
            warnings.warn("len on NamedTag is depreciated. Use the compound property to implement the same behaviour.", DeprecationWarning)
            return len(self.compound)

        def keys(self):
            """Depreciated. Do not use."""
            warnings.warn("keys on NamedTag is depreciated. Use the compound property to implement the same behaviour.", DeprecationWarning)
            return self.compound.keys()

        def values(self):
            """Depreciated. Do not use."""
            warnings.warn("values on NamedTag is depreciated. Use the compound property to implement the same behaviour.", DeprecationWarning)
            self.compound.values()

        def items(self):
            """Depreciated. Do not use."""
            warnings.warn("items on NamedTag is depreciated. Use the compound property to implement the same behaviour.", DeprecationWarning)
            return self.compound.items()

        def __setitem__(self, key: str, tag: AnyNBT):
            """Depreciated. Do not use."""
            warnings.warn("setitem on NamedTag is depreciated. Use the compound property to implement the same behaviour.", DeprecationWarning)
            self.compound[key] = tag

        def __delitem__(self, key: str):
            """Depreciated. Do not use."""
            warnings.warn("delitem on NamedTag is depreciated. Use the compound property to implement the same behaviour.", DeprecationWarning)
            del self.compound[key]

        def __contains__(self, key: str) -> bool:
            """Depreciated. Do not use."""
            warnings.warn("contains on NamedTag is depreciated. Use the compound property to implement the same behaviour.", DeprecationWarning)
            return key in self.compound

        def pop(self, k, default=None) -> AnyNBT:
            """Depreciated. Do not use."""
            warnings.warn("pop on NamedTag is depreciated. Use the compound property to implement the same behaviour.", DeprecationWarning)
            return self.compound.pop(k, default)

        def get(self, k, default=None) -> AnyNBT:
            """Depreciated. Do not use."""
            warnings.warn("get on NamedTag is depreciated. Use the compound property to implement the same behaviour.", DeprecationWarning)
            return self.compound.get(k, default)



from ._dtype import DecoderType, EncoderType

cdef class BufferContext:
    cdef char *buffer
    cdef readonly size_t size
    cdef readonly size_t offset

    cpdef bytes get_buffer(self)

cdef char *read_data(BufferContext buffer, size_t tag_size) except NULL
cdef void to_little_endian(void *data_buffer, int num_bytes, bint little_endian=*)

cdef char read_byte(BufferContext buffer) except? -1
cdef int read_int(BufferContext buffer, bint little_endian) except? -1
cdef str read_string(BufferContext buffer, bint little_endian, string_decoder: DecoderType)
cdef bytes read_bytes(BufferContext buffer, bint little_endian)

cdef void cwrite(object obj, char*buf, size_t length)
cdef void write_string(str s, object buffer, bint little_endian, string_encoder: EncoderType)
cdef void write_bytes(bytes b, object buffer, bint little_endian)
cdef void write_array(object value, object buffer, char size, bint little_endian)
cdef void write_byte(char value, object buffer)
cdef void write_short(short value, object buffer, bint little_endian)
cdef void write_int(int value, object buffer, bint little_endian)
cdef void write_long(long long value, object buffer, bint little_endian)
cdef void write_float(float value, object buffer, bint little_endian)
cdef void write_double(double value, object buffer, bint little_endian)

cpdef str utf8_decoder(bytes b)
cpdef bytes utf8_encoder(str s)
cpdef str utf8_escape_decoder(bytes b)
cpdef bytes utf8_escape_encoder(str s)



from ._value import AbstractBaseTag
from ._int import ByteTag, ShortTag, IntTag, LongTag
from ._float import FloatTag, DoubleTag
from ._string import StringTag
from ._list import ListTag
from ._compound import CompoundTag
from ._array import ByteArrayTag, IntArrayTag, LongArrayTag
from ._util import EncoderType
from mutf8 import encode_modified_utf8

class NamedTag:
    name: str
    tag: AbstractBaseTag
    def __init__(self, tag: AbstractBaseTag = None, name: str = ""): ...
    def to_snbt(self, indent=None, indent_chr=None) -> str: ...
    def to_nbt(
        self,
        *,
        compressed: bool = True,
        little_endian: bool = False,
        string_encoder: EncoderType = encode_modified_utf8,
    ) -> bytes: ...
    def save_to(
        self,
        filepath_or_buffer=None,
        *,
        compressed: bool = True,
        little_endian: bool = False,
        string_encoder: EncoderType = encode_modified_utf8,
    ) -> bytes: ...
    @property
    def byte(self) -> ByteTag: ...
    @property
    def short(self) -> ShortTag: ...
    @property
    def int(self) -> IntTag: ...
    @property
    def long(self) -> LongTag: ...
    @property
    def float(self) -> FloatTag: ...
    @property
    def double(self) -> DoubleTag: ...
    @property
    def string(self) -> StringTag: ...
    @property
    def list(self) -> ListTag: ...
    @property
    def compound(self) -> CompoundTag: ...
    @property
    def byte_array(self) -> ByteArrayTag: ...
    @property
    def int_array(self) -> IntArrayTag: ...
    @property
    def long_array(self) -> LongArrayTag: ...
